#!/usr/bin/env python
# -*- coding: utf-8 -*-
# generated by wxGlade 0.6 on Sun May 25 23:31:23 2008

# Copyright 2008 Martin Manns
# Distributed under the terms of the GNU General Public License
# generated by wxGlade 0.6 on Mon Mar 17 23:22:49 2008

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""
_dialogs
========

Provides:
---------
  1. CsvImportDialog: Dialog for CSV parameter choice
  2. MacroDialog: Dialog for macro management
  3. DimensionsEntryDialog
  4. AboutDialog
  
"""

import csv

import wx
import wx.lib.mixins.listctrl  as  listmix

from _widgets import IndexGrid, SortedListCtrl, MacroEditPanel
from _interfaces import CsvImport
from _datastructures import Macros
from config import DEBUG, VERSION, LIBPREFIX, icon_size, icons

class CsvImportDialog(wx.Dialog):
    """Dialog for CSV parameter choicewith preview grid
    
    Parameters:
    -----------
    csvfilepath: string, defaults to '.'
    \tPath and Filename of CSV input file
    
    """
    def __init__(self, *args, **kwds):
        try:
            self.csvfilepath = kwds.pop("csvfilepath")
        except KeyError:
            self.csvfilepath = '.'
        # begin wxGlade: CsvImportDialog.__init__
        kwds["style"] = wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER|wx.THICK_FRAME
        wx.Dialog.__init__(self, *args, **kwds)
        self.label_delimiter = wx.StaticText(self, -1, "Delimiter")
        self.text_ctrl_delimiter = wx.TextCtrl(self, -1, "")
        self.label_skipinitialspace = wx.StaticText(self, -1, "Skip initial space")
        self.checkbox_skipinitialspace = wx.CheckBox(self, -1, "")
        self.label_escapechar = wx.StaticText(self, -1, "Escape character")
        self.text_ctrl_escapechar = wx.TextCtrl(self, -1, "")
        self.label_doublequote = wx.StaticText(self, -1, "Doublequote")
        self.checkbox_doublequote = wx.CheckBox(self, -1, "")
        self.label_quotechar = wx.StaticText(self, -1, "Quote character")
        self.text_ctrl_quotechar = wx.TextCtrl(self, -1, "")
        self.panel_1 = wx.Panel(self, -1)
        self.panel_2 = wx.Panel(self, -1)
        self.csv_test_grid = IndexGrid(self, -1, size=(1, 1))
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "")
        self.button_apply = wx.Button(self, wx.ID_APPLY, "")
        self.button_ok = wx.Button(self, wx.ID_OK, "")

        self.__set_properties()
        self.__do_layout()

        self.Bind(wx.EVT_BUTTON, self.OnButtonApply, self.button_apply)
        # end wxGlade
        self.griddimension = (10, 10) 
        self.csv_test_grid.CreateGrid(self.griddimension[0], self.griddimension[1])
        self.csv_test_grid.SetRowLabelSize(0)
        self.csv_test_grid.SetColLabelSize(0)
        self.csv_test_grid.EnableEditing(0)
        self.csv_test_grid.EnableDragGridSize(0)
        self.dlgwidgets = {'delimiter': self.text_ctrl_delimiter, \
                           'skipinitialspace': self.checkbox_skipinitialspace, \
                           'escapechar': self.text_ctrl_escapechar, \
                           'doublequote': self.checkbox_doublequote, \
                           'quotechar': self.text_ctrl_quotechar}

    def __set_properties(self):
        # begin wxGlade: CsvImportDialog.__set_properties
        self.SetTitle("CSV Import")
        self.SetSize((410, 598))
        self.SetToolTipString("When True, whitespace immediately following the delimiter is ignored. The default is False.")
        self.label_delimiter.SetToolTipString("A one-character string used to separate fields. It defaults to ','.")
        self.text_ctrl_delimiter.SetMinSize((25, 25))
        self.text_ctrl_delimiter.SetToolTipString("A one-character string used to separate fields. It defaults to ','.")
        self.text_ctrl_delimiter.SetFocus()
        self.checkbox_skipinitialspace.SetToolTipString("When True, whitespace immediately following the delimiter is ignored. The default is False.")
        self.label_escapechar.SetToolTipString("A one-character string used by the writer to escape the delimiter if quoting is set to QUOTE_NONE and the quotechar if doublequote is False. On reading, the escapechar removes any special meaning from the following character. It defaults to None, which disables escaping.")
        self.text_ctrl_escapechar.SetMinSize((25, 25))
        self.text_ctrl_escapechar.SetToolTipString("A one-character string used by the writer to escape the delimiter if quoting is set to QUOTE_NONE and the quotechar if doublequote is False. On reading, the escapechar removes any special meaning from the following character. It defaults to None, which disables escaping.")
        self.label_doublequote.SetToolTipString("Controls how instances of quotechar appearing inside a field should be themselves be quoted. When True, the character is doubled. When False, the escapechar is used as a prefix to the quotechar. It defaults to True.")
        self.checkbox_doublequote.SetToolTipString("Controls how instances of quotechar appearing inside a field should be themselves be quoted. When True, the character is doubled. When False, the escapechar is used as a prefix to the quotechar. It defaults to True.")
        self.checkbox_doublequote.SetValue(1)
        self.label_quotechar.SetToolTipString("A one-character string used to quote fields containing special characters, such as the delimiter or quotechar, or which contain new-line characters. It defaults to '\"'.")
        self.text_ctrl_quotechar.SetMinSize((25, 25))
        self.text_ctrl_quotechar.SetToolTipString("A one-character string used to quote fields containing special characters, such as the delimiter or quotechar, or which contain new-line characters. It defaults to '\"'.")
        self.button_cancel.SetMinSize((80, 28))
        self.button_apply.SetMinSize((80, 28))
        self.button_ok.SetMinSize((80, 28))
        # end wxGlade

    def __do_layout(self):
        # begin wxGlade: CsvImportDialog.__do_layout
        grid_sizer_1 = wx.FlexGridSizer(2, 1, 0, 0)
        sizer_1 = wx.FlexGridSizer(1, 3, 5, 5)
        grid_sizer_2 = wx.FlexGridSizer(3, 4, 5, 5)
        grid_sizer_2.Add(self.label_delimiter, 0, wx.ADJUST_MINSIZE, 0)
        grid_sizer_2.Add(self.text_ctrl_delimiter, 0, wx.RIGHT|wx.ADJUST_MINSIZE, 3)
        grid_sizer_2.Add(self.label_skipinitialspace, 0, wx.LEFT|wx.ADJUST_MINSIZE, 15)
        grid_sizer_2.Add(self.checkbox_skipinitialspace, 0, wx.RIGHT|wx.ADJUST_MINSIZE, 3)
        grid_sizer_2.Add(self.label_escapechar, 0, wx.ADJUST_MINSIZE, 0)
        grid_sizer_2.Add(self.text_ctrl_escapechar, 0, wx.RIGHT|wx.ADJUST_MINSIZE, 3)
        grid_sizer_2.Add(self.label_doublequote, 0, wx.LEFT|wx.ADJUST_MINSIZE, 15)
        grid_sizer_2.Add(self.checkbox_doublequote, 0, wx.RIGHT|wx.ADJUST_MINSIZE, 3)
        grid_sizer_2.Add(self.label_quotechar, 0, wx.ADJUST_MINSIZE, 0)
        grid_sizer_2.Add(self.text_ctrl_quotechar, 0, wx.RIGHT|wx.ADJUST_MINSIZE, 3)
        grid_sizer_2.Add(self.panel_1, 1, wx.LEFT|wx.EXPAND, 15)
        grid_sizer_2.Add(self.panel_2, 1, wx.EXPAND, 0)
        grid_sizer_2.AddGrowableCol(3)
        grid_sizer_1.Add(grid_sizer_2, 1, wx.ALL|wx.EXPAND, 10)
        grid_sizer_1.Add(self.csv_test_grid, 1, wx.EXPAND, 0)
        sizer_1.Add(self.button_cancel, 0, wx.ALL|wx.EXPAND, 5)
        sizer_1.Add(self.button_apply, 0, wx.ALL|wx.EXPAND, 5)
        sizer_1.Add(self.button_ok, 0, wx.ALL|wx.EXPAND, 5)
        sizer_1.AddGrowableRow(0)
        sizer_1.AddGrowableCol(0)
        sizer_1.AddGrowableCol(1)
        sizer_1.AddGrowableCol(2)
        grid_sizer_1.Add(sizer_1, 1, wx.ALL|wx.EXPAND, 5)
        self.SetSizer(grid_sizer_1)
        grid_sizer_1.AddGrowableRow(1)
        grid_sizer_1.AddGrowableCol(0)
        self.Layout()
        self.Centre()
        # end wxGlade

    def OnButtonApply(self, event): # wxGlade: CsvImportDialog.<event_handler>
        if DEBUG: 
            print "Event handler 'OnButtonApply'"
        self.csv_test_grid.ClearGrid()
        csvimport = CsvImport(self.csvfilepath, self.get_csvargs())
        csvimport.fill_target(self.csv_test_grid)
        event.Skip()

    def get_csvargs(self):
        """Returns dict with csv import parameters from dialog."""
        replacements = {'\\t':'\t', '\\n':'\n', '\\b':'\b'}
        csvargs = {}
        for param in self.dlgwidgets:
            widget_value = self.dlgwidgets[param].GetValue()
            if widget_value != '':
                for char in replacements:
                    try: 
                        widget_value = widget_value.replace(char, replacements[char])
                        widget_value = widget_value.encode("utf-8")
                    except AttributeError:
                        pass
                csvargs[param] = widget_value
        csvargs['quoting'] = csv.QUOTE_NONNUMERIC
        return csvargs

# end of class CsvImportDialog


class MacroDialog(wx.Dialog, listmix.ColumnSorterMixin):
    """Macro management dialog"""
    def __init__(self, *args, **kwds):
        try:
            self.macros = Macros(kwds.pop("macros")) # Maps all macros names to the respective functions
        except KeyError:
            self.macros = Macros()

        # begin wxGlade: MacroDialog.__init__
        kwds["style"] = wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER|wx.THICK_FRAME
        wx.Dialog.__init__(self, *args, **kwds)
        self.window_1 = wx.SplitterWindow(self, -1, style=wx.SP_3D|wx.SP_BORDER)
        self.panel_3 = wx.Panel(self.window_1, -1)
        self.button_1 = wx.Button(self.panel_3, wx.ID_ADD, "")
        self.button_2 = wx.Button(self.panel_3, wx.ID_APPLY, "")
        self.button_3 = wx.Button(self.panel_3, wx.ID_REMOVE, "")
        self.macro_list_ctrl = SortedListCtrl(self.panel_3, -1, style=wx.LC_REPORT|wx.LC_SINGLE_SEL|wx.LC_SORT_ASCENDING|wx.SUNKEN_BORDER)
        self.MacroEditPanel = MacroEditPanel(self.window_1, -1)

        self.__set_properties()
        self.__do_layout()

        self.Bind(wx.EVT_BUTTON, self.OnAddMacro, self.button_1)
        self.Bind(wx.EVT_BUTTON, self.OnApplyChange, self.button_2)
        self.Bind(wx.EVT_BUTTON, self.OnRemoveMacro, self.button_3)
        self.Bind(wx.EVT_LIST_ITEM_FOCUSED, self.OnMacroFocused, self.macro_list_ctrl)
        # end wxGlade
        self.macro_list_ctrl.InsertColumn(0, "Function")
        self.macro_list_ctrl.InsertColumn(1, "Description")
#        self.Bind(wx.EVT_TEXT_ENTER, self.OnMacroTextEntered, self.MacroEditPanel.CodeTextCtrl)
        self.UpdateMacroList(func=None)

    def __set_properties(self):    
        # begin wxGlade: MacroDialog.__set_properties
        self.SetTitle("Macro list")
        self.SetSize((1122, 600))
        self.button_1.SetToolTipString("Add new macro")
        self.button_2.SetToolTipString("Apply changes to current macro")
        self.button_3.SetToolTipString("Remove current macro")
        self.MacroEditPanel.SetMinSize((-1, -1))
        # end wxGlade

    def __do_layout(self):
        # begin wxGlade: MacroDialog.__do_layout
        sizer_1 = wx.BoxSizer(wx.HORIZONTAL)
        grid_sizer_3 = wx.FlexGridSizer(2, 1, 5, 0)
        grid_sizer_4 = wx.FlexGridSizer(1, 3, 0, 5)
        grid_sizer_4.Add(self.button_1, 0, wx.ALL|wx.EXPAND, 3)
        grid_sizer_4.Add(self.button_2, 0, wx.ALL|wx.EXPAND, 3)
        grid_sizer_4.Add(self.button_3, 0, wx.ALL|wx.EXPAND, 3)
        grid_sizer_4.AddGrowableCol(0)
        grid_sizer_4.AddGrowableCol(1)
        grid_sizer_4.AddGrowableCol(2)
        grid_sizer_3.Add(grid_sizer_4, 1, wx.TOP|wx.EXPAND, 5)
        grid_sizer_3.Add(self.macro_list_ctrl, 1, wx.EXPAND, 0)
        self.panel_3.SetSizer(grid_sizer_3)
        grid_sizer_3.AddGrowableRow(1)
        grid_sizer_3.AddGrowableCol(0)
        self.window_1.SplitVertically(self.panel_3, self.MacroEditPanel, 500)
        sizer_1.Add(self.window_1, 1, wx.EXPAND, 0)
        self.SetSizer(sizer_1)
        self.Layout()
        # end wxGlade
        
    # Used by the ColumnSorterMixin, see wx/lib/mixins/listctrl.py
    def GetListCtrl(self):
        return self.macro_list_ctrl

    def GetCurrentMacro(self, event=None):
        """ Returns the macro that is selected in the list. None if there is no such macro"""
        if event is None:
            itemno = self.macro_list_ctrl.GetFocusedItem()
            if itemno == -1:
                current_listitem = None
            else:
                current_listitem = self.macro_list_ctrl.GetItem(itemno)
        else:
            current_listitem = self.macro_list_ctrl.GetItem(event.m_itemIndex)
        #print current_listitem
        try:
            return self.macros[current_listitem.GetText()]
        except KeyError:
            return None

    def GetMacroString(self):
        """ Gets the string-formatted call for the current function with the values from the macro entry form\nFormat: Functionname + ( + var1 + , +  ... + , + varn + )"""
        macro = self.GetCurrentMacro()
        functionname = macro.__name__
        varentries = [var.GetValue() for var in self.MacroEditPanel.varentries]
        varlist = varentries[:macro.func_code.co_argcount]
        return functionname + "(" + ", ".join(varlist) + ")"

    def UpdateMacroList(self, func=None):
        """ Updates the macro list from self.macros """
        self.macro_list_ctrl.DeleteAllItems()
        for i, key in enumerate(sorted(list(self.macros))):
            listentry = (key, self.macros[key].__doc__)
            index = self.macro_list_ctrl.Append(listentry)
            self.macro_list_ctrl.SetItemData(index, i)
        if func is not None:
            self.UpdateMacroListSelection(func)

    def UpdateMacroListSelection(self, func):
        """ Updates the selection so that it focuses
            the item of the recently selected macro """
        newitempos = self.macro_list_ctrl.FindItem(-1, func.__name__)
        self.macro_list_ctrl.Select(newitempos)        

    def OnAddMacro(self, event): # wxGlade: MacroDialog.<event_handler>
        code = self.MacroEditPanel.CodeTextCtrl.GetText()
        func = self.macros.AddToMacroDict(code)
        if func != 0:
            self.UpdateMacroList(func)
        event.Skip()

    def OnApplyChange(self, event): # wxGlade: MacroDialog.<event_handler>
        """Applies change in function to macro dict"""
        code = self.MacroEditPanel.CodeTextCtrl.GetText()
        func = self.macros.GetMacro(code)
        if func.__name__ not in self.macros:
            # If the function name has changed, the old function is removed            
            selecteditem_pos = self.macro_list_ctrl.GetFirstSelected()
            if selecteditem_pos != -1:
                current_listitem = self.macro_list_ctrl.GetItem(selecteditem_pos)
                current_macro = self.macros[current_listitem.GetText()]
                self.macros.pop(current_macro.__name__)
        self.macros[func.__name__] = func
        self.UpdateMacroList(func)
        #for variable in func.func_code.co_varnames:
        #    print variable
        event.Skip()

    def OnRemoveMacro(self, event): # wxGlade: MacroDialog.<event_handler>
        if DEBUG: print "Event handler `OnRemoveMacro'"
        selecteditem_pos = self.macro_list_ctrl.GetFirstSelected()
        if selecteditem_pos != -1:
            current_listitem = self.macro_list_ctrl.GetItem(selecteditem_pos)
            current_macro = self.macros[current_listitem.GetText()]
            self.macros.pop(current_macro.__name__)
            self.UpdateMacroList()
        event.Skip()

    def OnMacroFocused(self, event): # wxGlade: MacroDialog.<event_handler>
        if DEBUG: print "Event handler `OnMacroFocused'"
        current_macro = self.GetCurrentMacro(event)
        if current_macro is None:
            event.Skip()
            return 0
        self.MacroEditPanel.CodeTextCtrl.SetText(current_macro.func_dict['macrocode'])
        try:
            self.MacroEditPanel.docstringTextCtrl.SetValue(current_macro.__doc__)
        except TypeError:
            self.MacroEditPanel.docstringTextCtrl.SetValue("")
        self.MacroEditPanel.update_macroform(self.MacroEditPanel.pane_macroform, current_macro)
        event.Skip()

# end of class MacroDialog


class DimensionsEntryDialog(wx.Dialog):
    """Input dialog for the 3 dimensions of a grid"""
    def __init__(self, *args, **kwds):
        # begin wxGlade: DimensionsEntryDialog.__init__
        kwds["style"] = wx.DEFAULT_DIALOG_STYLE|wx.MINIMIZE_BOX|wx.STAY_ON_TOP
        wx.Dialog.__init__(self, *args, **kwds)
        self.Rows_Label = wx.StaticText(self, -1, "Rows", style=wx.ALIGN_CENTRE)
        self.X_DimensionsEntry = wx.TextCtrl(self, -1, "")
        self.Columns_Label = wx.StaticText(self, -1, "Columns", style=wx.ALIGN_CENTRE)
        self.Y_DimensionsEntry = wx.TextCtrl(self, -1, "")
        self.Tabs_Label = wx.StaticText(self, -1, "Tabs", style=wx.ALIGN_CENTRE)
        self.Z_DimensionsEntry = wx.TextCtrl(self, -1, "")
        self.cancel_button = wx.Button(self, wx.ID_CANCEL, "")
        self.ok_button = wx.Button(self, wx.ID_OK, "")

        self.__set_properties()
        self.__do_layout()

        self.Bind(wx.EVT_TEXT, self.OnXDim, self.X_DimensionsEntry)
        self.Bind(wx.EVT_TEXT, self.OnYDim, self.Y_DimensionsEntry)
        self.Bind(wx.EVT_TEXT, self.OnZDim, self.Z_DimensionsEntry)
        # end wxGlade
        self.dimensions = [100, 100, 1]

    def __set_properties(self):
        # begin wxGlade: DimensionsEntryDialog.__set_properties
        self.SetTitle("Create a new grid and discard the old one")
        self.cancel_button.SetDefault()
        # end wxGlade

    def __do_layout(self):
        # begin wxGlade: DimensionsEntryDialog.__do_layout
        grid_sizer_1 = wx.GridSizer(4, 2, 3, 3)
        grid_sizer_1.Add(self.Rows_Label, 0, wx.LEFT|wx.ALIGN_CENTER_VERTICAL, 3)
        grid_sizer_1.Add(self.X_DimensionsEntry, 0, wx.EXPAND, 0)
        grid_sizer_1.Add(self.Columns_Label, 0, wx.LEFT|wx.ALIGN_CENTER_VERTICAL, 3)
        grid_sizer_1.Add(self.Y_DimensionsEntry, 0, wx.EXPAND, 0)
        grid_sizer_1.Add(self.Tabs_Label, 0, wx.LEFT|wx.ALIGN_CENTER_VERTICAL, 3)
        grid_sizer_1.Add(self.Z_DimensionsEntry, 0, wx.EXPAND, 0)
        grid_sizer_1.Add(self.cancel_button, 0, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_HORIZONTAL|wx.ALIGN_CENTER_VERTICAL|wx.FIXED_MINSIZE, 3)
        grid_sizer_1.Add(self.ok_button, 0, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_HORIZONTAL|wx.ALIGN_CENTER_VERTICAL|wx.FIXED_MINSIZE, 3)
        self.SetSizer(grid_sizer_1)
        grid_sizer_1.Fit(self)
        self.Layout()
        # end wxGlade

    def OnXDim(self, event): # wxGlade: DimensionsEntryDialog.<event_handler>
        try: self.dimensions[0] = int(event.GetString())
        except ValueError: self.dimensions[0] = 0
        event.Skip()

    def OnYDim(self, event): # wxGlade: DimensionsEntryDialog.<event_handler>
        try: self.dimensions[1] = int(event.GetString())
        except ValueError: self.dimensions[1] = 0
        event.Skip()

    def OnZDim(self, event): # wxGlade: DimensionsEntryDialog.<event_handler>
        try: self.dimensions[2] = int(event.GetString())
        except ValueError: self.dimensions[2] = 0
        event.Skip()

# end of class DimensionsEntryDialog


class AboutDialog(wx.Dialog):
    """Displays information about pyspread"""
    def __init__(self, *args, **kwds):
        # begin wxGlade: AboutDialog.__init__
        kwds["style"] = wx.DEFAULT_DIALOG_STYLE
        wx.Dialog.__init__(self, *args, **kwds)
        self.bitmap_2 = wx.StaticBitmap(self, -1, wx.Bitmap(LIBPREFIX+'icons/pyspread.png', wx.BITMAP_TYPE_ANY))
        self.about_label = wx.StaticText(self, -1, "", style=wx.ALIGN_CENTRE)

        self.__set_properties()
        self.__do_layout()
        # end wxGlade

    def __set_properties(self):
        # begin wxGlade: AboutDialog.__set_properties
        self.SetTitle("About pyspread")
        # end wxGlade
        
        self.about_label.SetLabel("pyspread " + VERSION + \
                                  "\nCopyright Martin Manns 2008")

    def __do_layout(self):
        # begin wxGlade: AboutDialog.__do_layout
        sizer_4 = wx.BoxSizer(wx.HORIZONTAL)
        sizer_4.Add(self.bitmap_2, 0, wx.ALL|wx.ALIGN_CENTER_HORIZONTAL|wx.ALIGN_CENTER_VERTICAL, 10)
        sizer_4.Add(self.about_label, 0, wx.ALL|wx.ALIGN_CENTER_HORIZONTAL|wx.ALIGN_CENTER_VERTICAL, 10)
        self.SetSizer(sizer_4)
        sizer_4.Fit(self)
        self.Layout()
        self.Centre()
        # end wxGlade

# end of class AboutDialog



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# generated by wxGlade 0.6 on Sun May 25 23:31:23 2008

# Copyright 2008 Martin Manns
# Distributed under the terms of the GNU General Public License
# generated by wxGlade 0.6 on Mon Mar 17 23:22:49 2008

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""
_choicebars
===========

Provides choice bars such as toolbars and menubars

Provides:
---------
  1. ContextMenu: Context menu for main grid
  2. MainMenu: Main menu of pyspread
  3. MainToolbar: Main toolbar of pyspread

"""
import wx

from config import icons

class _filledMenu(wx.Menu):
    """Menu that fills from the attribute menudata.
    
    Parameters:
    parent: object
    \tThe parent object that hosts the event handler methods
    menubar: wx.Menubar, defaults to parent
    \tThe menubar to which the menu is attached
    
    menudata has the following structure:
    [
        [wx.Menu, "Menuname", [\
            [wx.MenuItem, ["Methodname", "Itemlabel", "Help"]] , \
            ...
            "Separator" , \
            ...
            [wx.Menu, ...], \
            ...
        ] , \
    ...
    ]
    """
    
    menudata = []
    
    def _add_submenu(self, parent, data):
        """Adds items in data as a submenu to parent"""
        
        for item in data:
            obj = item[0]
            if obj == wx.Menu:
                menuname = item[1]
                submenu = item[2]
                menu = obj()
                self._add_submenu(menu, submenu)
                self.menubar.Append(menu, menuname)
            elif obj == wx.MenuItem:
                methodname = item[1][0]
                method = self.parent.__getattribute__(methodname)
                params = [parent, wx.NewId()] + item[1][1:] + [wx.ITEM_NORMAL]
                menuitem = obj(*params)
                parent.AppendItem(menuitem)
                self.parent.Bind(wx.EVT_MENU, method, menuitem)
            elif obj == "Separator":
                parent.AppendSeparator()
            else:
                raise TypeError, "Menu item unknown"


    def __init__(self, *args, **kwargs):
        self.parent = kwargs.pop('parent')
        try:
            self.menubar = kwargs.pop('menubar')
        except KeyError:
            self.menubar = self.parent
        wx.Menu.__init__(self, *args, **kwargs)
        self._add_submenu(self, self.menudata)

# end of class _filledMenu


class ContextMenu(_filledMenu):
    """Context menu for grid operations"""
    
    item = wx.MenuItem
    menudata = [ \
    [item, ["OnCut", "Cu&t\tCtrl+x", "Cut cell to clipboard"]], \
    [item, ["OnCopy", "&Copy\tCtrl+c", "Copy input strings to clipboard"]], \
    [item, ["OnPaste", "&Paste\tCtrl+v", "Paste cell from clipboard"]], \
    [item, ["OnInsertRows", "Insert &rows\tShift+Ctrl+i", "Insert rows at cursor"]], \
    [item, ["OnInsertColumns", "&Insert columns\tCtrl+i", "Insert columns at cursor"]], \
    [item, ["OnDeleteRows", "Delete rows\tShift+Ctrl+d", "Delete rows" ]], \
    [item, ["OnDeleteColumns", "Delete columns\tCtrl+Alt+d", "Delete columns"]]]
    
            
# end of class ContextMenu


class MainMenu(_filledMenu):
    """Main application menu"""
    item = wx.MenuItem
    menudata = [ \
        [wx.Menu, "&File", [\
            [item, ["OnFileNew", "&New\tCtrl+n", "Create a new, empty spreadsheet"]], \
            [item, ["OnFileOpen", "&Open\tCtrl+o", "Open spreadsheet from file"]], \
            [item, ["OnFileSave", "&Save\tCtrl+s", "Save spreadsheet to file"]], \
            ["Separator"], \
            [item, ["OnCSVImport", "CSV &Import", "Import a CSV format file"]], \
            ["Separator"], \
            [item, ["OnFilePrint", "&Print\tCtrl+p", "Print current spreadsheet"]], \
            ["Separator"], \
            [item, ["OnExit", "E&xit\tCtrl+q", "Exit Program"]]] \
        ], \
        [wx.Menu, "&Edit", [\
            [item, ["OnUndo", "&Undo\tCtrl+z", "Undo last step"]], \
            [item, ["OnRedo", "&Redo\tShift+Ctrl+z", "Redo last undone step"]], \
            ["Separator"], \
            [item, ["OnCut", "Cu&t\tCtrl+x", "Cut cell to clipboard"]], \
            [item, ["OnCopy", "&Copy\tCtrl+c", "Copy the input strings of the cells to clipboard"]], \
            [item, ["OnCopyResult", "Copy &Results\tShift+Ctrl+c", "Copy the result strings of the cells to the clipboard"]], \
            [item, ["OnPaste", "&Paste\tCtrl+v", "Paste cells from clipboard"]], \
            ["Separator"], \
            [item, ["OnShowFind", "&Find\tCtrl+f", "Find cell by content"]], \
            [item, ["OnShowFindReplace", "Replace\tCtrl+Shift+f", "Replace strings in cells"]], \
            ["Separator"], \
            [item, ["OnInsertRows", "Insert &rows\tShift+Ctrl+i", "Insert rows at cursor"]], \
            [item, ["OnInsertColumns", "&Insert columns\tCtrl+i", "Insert columns at cursor"]], \
            [item, ["OnInsertTable", "Insert &table\tShift+Ctrl+Alt+i", "Insert table before current table"]], \
            ["Separator"], \
            [item, ["OnDeleteRows", "Delete rows\tShift+Ctrl+d", "Delete rows"]], \
            [item, ["OnDeleteColumns", "Delete columns\tCtrl+Alt+d", "Delete columns"]], \
            [item, ["OnDeleteTable", "Delete table\tShift+Ctrl+Alt+d", "Delete current table"]]] \
        ], \
        [wx.Menu, "&Macro", [\
            [item, ["OnMacroList", "&Macro list\tCtrl+m", "Choose, fill in, manage, and create macros"]], \
            [item, ["OnMacroListLoad", "&Load macro list\tShift+Ctrl+m", "Load macro list"]], \
            [item, ["OnMacroListSave", "&Save macro list\tShift+Alt+m", "Save macro list"]]] \
        ], \
        [wx.Menu, "&Help", [\
            [item, ["OnAbout", "&About", "About this program"]]] \
        ] \
    ]

# end of class MainMenu


class MainToolbar(wx.ToolBar):
    """Main application toolbar, built from attribute toolbardata
    
    toolbardata has the following structure:
    [[toolobject, "Methodname", "Label", 
                  "Iconname", "Tooltip", "Help string"] , \
    ...
    ["Separator"] ,\
    ...
    ]
    
    """
    
    tool = wx.ITEM_NORMAL
    
    toolbardata = [
    [tool, "OnFileNew", "New", "FileNew", "New spreadsheet", "Create a new, empty spreadsheet"], \
    [tool, "OnFileOpen", "Open", "FileOpen", "Open spreadsheet", "Open spreadsheet from file"], \
    [tool, "OnFileSave", "Save", "FileSave", "Save spreadsheet", "Save spreadsheet to file"], \
    ["Separator"] , \
    [tool, "OnUndo", "Undo", "Undo", "Undo", "Undo last operation"], \
    [tool, "OnRedo", "Redo", "Redo", "Redo", "Redo next operation"], \
    ["Separator"] , \
    [tool, "OnShowFind", "Find", "Find", "Find", "Find cell by content"], \
    [tool, "OnShowFindReplace", "Replace", "FindReplace", "Replace", "Replace strings in cells"], \
    ["Separator"] , \
    [tool, "OnCut", "Cut", "EditCut", "Cut", "Cut cells to clipboard"], \
    [tool, "OnCopy", "Copy", "EditCopy", "Copy", "Copy the input strings of the cells to clipboard"], \
    [tool, "OnCopyResult", "Copy Results", "EditCopyRes", "Copy Results", "Copy the result strings of the cells to the clipboard"], \
    [tool, "OnPaste", "Paste", "EditPaste", "Paste", "Paste cell from clipboard"], \
    ["Separator"] , \
    [tool, "OnFilePrint", "Print", "FilePrint", "Print current spreadsheet", "Print current spreadsheet"], \
    ]
    
    def _add_tools(self):
        """Adds tools from self.toolbardata to self"""
        for tool in self.toolbardata:
            obj = tool[0]
            if obj == "Separator":
                self.AddSeparator()
            elif obj == self.tool:
                methodname = tool[1]
                method = self.parent.__getattribute__(methodname)
                label = tool[2]
                icon = wx.Bitmap(icons[tool[3]], wx.BITMAP_TYPE_ANY)
                icon2 = wx.NullBitmap
                tooltip = tool[4]
                helpstring = tool[5]
                toolid = wx.NewId()
                self.AddLabelTool(toolid, label, icon, icon2, obj, tooltip, helpstring)
                self.parent.Bind(wx.EVT_TOOL, method, id=toolid)
            else:
                raise TypeError, "Toolbar item unknown"
    
    def __init__(self, *args, **kwargs):
        wx.ToolBar.__init__(self, *args, **kwargs)
        self.parent = args[0]
        self._add_tools()
        

# end of class MainToolbar



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2008 Martin Manns
# Distributed under the terms of the GNU General Public License
# generated by wxGlade 0.6 on Mon Mar 17 23:22:49 2008

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""
_datastructures
===============

Provides
  1. Grid object class Grid
  2. Macro user dictionary class Macros
  3. Undo/Redo framework class UnRedo

"""

import types
import UserDict
from itertools import izip

import numpy

from config import DEBUG
from _interfaces import string_match

X, Y, Z, S = [None] * 4

class PyspreadGrid(object):
    """Central data object that provides two numpy based 3D object arrays.

    One array (sgrid) stores strings that contain python expressions. The 
    second array (fgrid) stores functions that are derived from these strings.
    This class makes use of eval and does not provide any security features.
    It furthermore stores functions as globals if a function name is given in 
    the expression (e.g. 'a = 2' --> a = lambda x=0: 2; global a).

    Parameters
    ----------
    parent: object
    \tParent class
    dimensions : 3 tuple of int
    \tThe dimensions of sgrid and fgrid (defaults to (10, 10, 10))

    """
        
    def __init__(self, dimensions=(10, 10, 10)):
        global S
        S = self
        self.isrecursion = 0
        self.name = "S"
        self.fgrid = numpy.zeros(dimensions, dtype="O")
        self.sgrid = numpy.zeros(dimensions, dtype="S")
        self.sgrid = numpy.array(self.sgrid, dtype="O")
        self.macros = Macros({}) # Macros from Macrolist
        self.unredo = UnRedo()
        self.resultcache = {}

    def __getitem__(self, pos):
        if repr(pos) in self.resultcache:
            return self.resultcache[repr(pos)]
        global X, Y, Z
        _pos_storage = X, Y, Z
        X, Y, Z = pos
        self.isrecursion += 1
        try:
            fgrid_pos = self.fgrid[pos]
        except IndexError, error_message: 
            self.isrecursion -= 1
            return error_message
        if type(fgrid_pos) is types.FunctionType:
            fgrid_pos = numpy.array([fgrid_pos], dtype="O")
        if fgrid_pos is 0:
            self.isrecursion -= 1
            return ''
        
        # We are now sure that fgrid_pos is a kind of list here
        if type(fgrid_pos) in [types.TupleType, types.ListType]:
            fgrid_pos = numpy.array(fgrid_pos, dtype = "O")
        elif type(fgrid_pos) is not numpy.ndarray:
            raise AssertionError, "fgrid_pos has wrong type: " + \
                                  str(type(fgrid_pos))
        # We are now sure that fgrid_pos is a numpy.ndarray
        #print fgrid_pos
        res = self.__call_fgrid_function(fgrid_pos)
        X, Y, Z = _pos_storage
        self.isrecursion -= 1
        if repr(pos) not in self.resultcache:
            self.resultcache[repr(pos)] = res
        return res
    
    def __setitem__(self, pos, value):
        self.resultcache = {}
        undo_operation = (self.__setitem__, [pos, self.sgrid.__getitem__(pos)])
        redo_operation = (self.__setitem__, [pos, value])
        self.unredo.append(undo_operation, redo_operation)
        self.sgrid.__setitem__(pos, value)
        return value

    def __len__(self):
        return len(self.sgrid)

    def __call_fgrid_function(self, fgrid_pos):
        """
        Calls functions in the numpy.ndarray fgrid_pos and returns result
        
        Parameters:
        -----------
        fgrid_pos: Function array that is called
        
        """
        
        global X, Y, Z
        _pos_orig = X, Y, Z
        
        def _call(func, pos):
            if func is 0:
                return 0
            else:
                try:
                    return func(pos)
                except Exception, error_message:
                    return error_message
        
        try: 
            func_it = fgrid_pos.flat
        except:
            raise AssertionError, str(type(fgrid_pos))+" is not numpy.ndarray"
        resultlist = []
        for i, f in enumerate(func_it):
            #print self.getshapedpos(fgrid_pos, i)
            shapedpos = self.getshapedpos(fgrid_pos, i)
            #print _pos_orig, shapedpos
            try:
                X = _pos_orig[0].start + shapedpos[0]
                Y = _pos_orig[1].start + shapedpos[1]
                Z = _pos_orig[2].start + shapedpos[2]
            except:
                pass
            resultlist.append(_call(f, pos=i))
        result = numpy.array(resultlist)
        try:
            len(result)
        except TypeError:
            return result
        if len(result) == 1:
            return result[0]
        else:
            return result.reshape(fgrid_pos.shape)

    def insert(self, insertionpoint, notoinsert, newcells=None):
        """Insert rows, columns or tables
        
        Parameters:
        -----------
        insertionpoint: 3-tuple or list with 3 elements
        \tinsertionpoint must be a 3 tuple, which is None except for one element.
        \tThis element corresponds to the dimension of the insertion operation
        \tand describes the position of the insertion operation.
        notoinsert: int
        \tNo. cols/rows/tables that are to be inserted
        
        """
        
        # The function does only work correctly with correct insertionpoint
        assert len(insertionpoint) == 3
        assert list(insertionpoint).count(None) == 2
        
        undo_operation = (self.remove, [insertionpoint, notoinsert])
        redo_operation = (self.insert, [insertionpoint, notoinsert, newcells])
        self.unredo.append(undo_operation, redo_operation)
        
        ip = list(insertionpoint)
        axis = ip.index(max(ip))
        if newcells is None:
            newshapeindices = list(self.sgrid.shape)
            newshapeindices[axis] = notoinsert
            newcells = numpy.zeros(newshapeindices, dtype="O")
        
        def _subarrays(array, newcells, key):
            """Returns tuple of sub-arrays that can be concatenated"""
            subarrays = (array[:key[0], :key[1], :key[2]], \
                         newcells, \
                         array[key[0]:, key[1]:, key[2]:])
            return filter(lambda x:list(x.flat) != [], subarrays)
            
        subarrays = _subarrays(self.sgrid, newcells, ip)
        self.sgrid = numpy.concatenate(subarrays, axis=axis)
        subarrays = _subarrays(self.fgrid, newcells, ip)
        self.fgrid = numpy.concatenate(subarrays, axis=axis)

    def getflatpos(self, array, pos):
        """Get index of a cell in a flattened numpy array in Fortran order."""
        from operator import mul
        cumgsr = [1] + \
            [reduce(mul, array.shape[:i]) for i in xrange(1, len(array.shape))]
        return sum(p * e for p, e in izip(pos, cumgsr))
        
    def getshapedpos(self, array, pos):
        """Get cell index from a flattened array index in Fortran order."""
        shape = array.shape
        def _rek(i, pos):
            try: currpos = pos % shape[i]
            except IndexError: return []
            return [currpos] + _rek(i+1, pos // shape[i])
        return tuple(_rek(0, pos))

    def remove(self, removalpoint, notoremove):
        """Remove rows, columns or tables
                
        Parameters:
        -----------
        removalpoint: 3-tuple or list with 3 elements
        \removalpoint must be a 3 tuple, which is None except for one element.
        \tThis element corresponds to the dimension of the removal operation
        \tand describes the position of the removal operation.
        notoremove: int
        \tNo. cols/rows/tables that are to be removed
        
        """
        
        rp = removalpoint
        axis = rp.index(max(rp))
        removedcellslice = [slice(rpele, rpele) for rpele in rp]
        removedcellslice[axis] = slice(removedcellslice[axis].start, \
                                       removedcellslice[axis].stop + notoremove)
        removedcells = self.sgrid[removedcellslice]
        
        undo_operation = (self.insert, [removalpoint, notoremove, removedcells])
        redo_operation = (self.remove, [removalpoint, notoremove])
        self.unredo.append(undo_operation, redo_operation)

        self.sgrid = numpy.delete(self.sgrid, removedcellslice[axis], axis)
        self.fgrid = numpy.delete(self.fgrid, removedcellslice[axis], axis)

        
    def spread(self, value, pos):
        """Spread values into the grid with the top-left-upmost position x, y, z
        
        Parameters
        ----------
        value: Scalar, iterable or array
        \tValues that shall be spread
        pos: 3-tuple of int
        \tValid index of self for top-left value insertion
        
        """
        
        x, y, z = pos
        valdim = numpy.ndim(value)
        if valdim in xrange(1, 4):
            value = numpy.array(value)
            shape = value.shape
            
            if valdim == 1:
                rowsleft = self.sgrid[x:, :, :].shape[0] - shape[0]
                if rowsleft < 0:
                    value = value[:shape[0] + rowsleft]
            if valdim == 2:
                rowsleft = self.sgrid[x:, :, :].shape[0] - shape[0]
                colsleft = self.sgrid[:, y:, :].shape[1] - shape[1]
                if rowsleft < 0:
                    value = value[:shape[0] + rowsleft, :]
                if colsleft < 0:
                    value = value[:, :shape[1] + colsleft]
            if valdim > 2:
                rowsleft = self.sgrid[x:, :, :].shape[0] - shape[0]
                colsleft = self.sgrid[:, y:, :].shape[1] - shape[1]
                tablesleft = self.sgrid[:, :, z:].shape[2] - shape[2]
                if rowsleft < 0:
                    value = value[:shape[0] + rowsleft, :, :]
                if colsleft < 0:
                    value = value[:, :shape[1] + colsleft, :]
                if tablesleft < 0:
                    value = value[:, :, :shape[2] + tablesleft]
            
            strings = numpy.array(map(repr, value.flat)).reshape(value.shape)
            
            if valdim == 1:
                S[x: x + shape[0], y, z] = strings
            elif valdim == 2:
                S[x: x + shape[0], y: y + shape[1], z] = strings
            elif valdim == 3:
                S[x: x + shape[0], y: y + shape[1], z: z + shape[2]] = strings
        elif valdim == 0:
            #res = self.__setitem__(slice(target, target, None), repr(value))
            S[x, y, z] = repr(value)
        else:
            raise ValueError, "Dimension of " + str(value) + " too high"
        self.eval_function_grid()

    def findnextmatch(self, startpos, find_string, flags):
        """ Returns a tuple with the position of the next match of find_string
        
        Returns None if string not found.
        
        Parameters:
        -----------
        startpos:   Start position of search
        find_string:String to be searched for
        flags:      List of strings, out ouf 
                    ["UP" xor "DOWN", "WHOLE_WORD", "MATCH_CASE"]
        
        """
        #print repr(find_string)
        # Index of current cell in flattened grid
        flatgridpos = self.getflatpos(self.sgrid, startpos) 
        # Search in each cell until match
        flatgrid = self.sgrid.flatten('F') # We need Fortran order here
        # Start at flatgridpos
        searchgrid = numpy.concatenate((flatgrid[flatgridpos:], \
                                        flatgrid[:flatgridpos]))
        if "UP" in flags:
            searchgrid = searchgrid[::-1]
        findpos = None
        for key, datastring in enumerate(searchgrid):
            if datastring != '' and \
               string_match(datastring, find_string, flags) is not None:
                if "DOWN" in flags:
                    correctedpos =  flatgridpos + key
                elif "UP" in flags:
                    correctedpos =  flatgridpos - key - 1
                else:
                    raise AttributeError, "UP xor DOWN flag not present in " + \
                                          flags
                findpos = self.getshapedpos(self.sgrid, correctedpos)
                break
        return findpos

    def get_function_cell_indices(self, gridslice = None):
        """
        Get indices for all fgrid cells != 0
        
        Parameters:
        -----------
        gridslice: 3-tuple of slices
        \tThe slice of the grid that is searched. Defaults to whole grid.
        
        """
        if gridslice is None:
            gridslice = tuple([slice(None)]*len(self.sgrid.shape))
        sgrid = self.sgrid[gridslice]
        fgrid = self.fgrid[gridslice]
        nonzero = list(set(zip(*numpy.nonzero(sgrid)) + \
                      zip(*numpy.nonzero(fgrid))))
        gridslice = list(gridslice)
        for i in xrange(len(self.sgrid.shape)):
            if gridslice[i].start is None:
                gridslice[i] = slice(0, gridslice[i].stop, gridslice[i].step)
        nonzero = [tuple(cell[i] + gridslice[i].start \
                        for i in xrange(len(self.sgrid.shape))) \
                            for cell in nonzero]
        return nonzero

    def eval_function_grid(self, gridslice = None):
        """ 
        Eval the function grid from the string grid 
        
        Parameters:
        -----------
        gridslice: 3-tuple of slices
        \tThe slice of the grid that is searched. Defaults to whole grid.\
        
        """
        if gridslice is None:
            gridslice = tuple([slice(None, None)]*len(self.sgrid.shape))
        sgrid = self.sgrid[gridslice]
        fgrid = self.fgrid[gridslice]
        string_elements = set(zip(*numpy.nonzero(sgrid)) + \
                              zip(*numpy.nonzero(fgrid)))
        global S
        S = self
        for i in string_elements:
            err = None 
            if sgrid[i] == "":
                fgrid[i] = 0
            else:
                try:
                    iter(sgrid[i])
                except TypeError:
                    raise TypeError, "sgrid may only hold iterable strings." + \
                                     "sgrid[" + str(i) + "] is " + \
                                     str(type(sgrid[i])) + \
                                     " that is not iterable."
                try:
                    expression = "".join(["lambda pos:", sgrid[i]])
                    fgrid[i] = eval(expression, globals())
                except Exception, x:
                    err = x
            # We got an error that may be caused by a global assignment ("a=2")
            if err is not None and '=' in sgrid[i]:
                try:
                    stripped_string = sgrid[i].split('=', 1)
                    if stripped_string[0] in ['S', 'X', 'Y', 'Z']:
                        raise SyntaxError, \
                              'Cell name must not mask built-in global.'
                    exec("=".join(stripped_string[:2])) in globals()
                    exec("".join(["global ", stripped_string[0]])) in globals()
                    evalstring = "".join(["lambda pos:", stripped_string[1]])
                    fgrid[i] = eval(evalstring, globals())
                    err = None
                except Exception, x:
                    err = x
            if err is not None:
                fgrid[i] = lambda x: err

    def set_global_macros(self, macros=None):
        """ Sets macros to global scope """
        
        self.resultcache = {}
        if macros is None: 
            macros = self.macros
        for macroname, macro in macros.iteritems():
            globals()[macroname] = macro

# end of class PyspreadGrid


class Macros(UserDict.IterableUserDict):
    """User dict class for macros.

    This class provides a getter and setter method for storing the full
    macro Python code in the 'macrocode' attribute of the funcdict.

    """
    def GetMacro(self, code):
        """ Returns the function derived from the code string code. """
        funcname = code.split("(")[0][3:].strip()
        # Windows exec does not like Windows newline
        code = code.replace('\r\n', '\n')
        exec(code)
        func = eval(funcname, globals(), locals())
        func.func_dict['macrocode'] = code
        return func
        
    def AddToMacroDict(self, code):
        """ Adds a macro with the code string 'code' to the macro dict"""
        func = self.GetMacro(code)
        if func.__name__ in self:
            if DEBUG: print "Macro already present."
            return 0
        self[func.__name__] = func
        return func
        
# end of class Macros

class UnRedo(object):
    """Undo/Redo framework class.
    
    For each undoable operation, the undo/redo framework stores the
    undo operation and the redo operation. For each step, a 4-tuple of:
    1) the function object that has to be called for the undo operation
    2) the undo function paarmeters as a list
    3) the function object that has to be called for the redo operation
    4) the redo function paarmeters as a list
    is stored.
    
    One undo step in the application can comprise of multiple operations.
    Undo steps are separated by the string "MARK".
    
    The attributes should only be written to by the class methods.

    Attributes
    ----------
    undolist: List
    \t
    redolist: List
    \t
    active: Boolean
    \tTrue while an undo or a redo step is executed.

    """
    
    def __init__(self):
        # [(undofunc, [undoparams, ...], redofunc, [redoparams, ...]), 
        # ..., "MARK", ...]
        # "MARK" separartes undo/redo steps
        self.undolist = []
        self.redolist = []
        self.active = False
        
    def mark(self):
        """Inserts a mark in undolist and empties redolist"""
        if self.undolist != [] and self.undolist[-1] != "MARK":
            self.undolist.append("MARK")
        self.redolist = []
    
    def undo(self):
        """Undos operations until next mark and stores them in the redolist"""
        self.active = True
        while self.undolist != [] and self.undolist[-1] == "MARK":
            self.undolist.pop()
        if self.redolist != [] and self.redolist[-1] != "MARK":
            self.redolist.append("MARK")
        while self.undolist != []:
            step = self.undolist.pop()
            if step == "MARK": 
                break
            self.redolist.append(step)
            step[0](*step[1])
        self.active = False
        
    def redo(self):
        """Redos operations until next mark and stores them in the undolist"""
        self.active = True
        while self.redolist != [] and self.redolist[-1] == "MARK":
            self.redolist.pop()
        if self.undolist != [] and self.undolist[-1] != "MARK":
            self.undolist.append("MARK")        
        while self.redolist != []:
            step = self.redolist.pop()
            if step == "MARK": 
                break
            self.undolist.append(step)
            step[2](*step[3])
        self.active = False

    def reset(self):
        """Empties both undolist and redolist"""
        self.__init__()

    def append(self, undo_operation, operation):
        """Stores an operation and its undo operation in the undolist"""
        if not self.active:
            self.undolist.append(undo_operation + operation)
            #print undo_operation + operation

# end of class UnRedo



"""
pyspread config file
====================

"""

import wx
import distutils.sysconfig

LIBPREFIX = distutils.sysconfig.get_python_lib() + '/pyspread/'
try:
    f = open(LIBPREFIX + 'pyspread.py', 'r')
    f.close()
except IOError:
    LIBPREFIX = './'

DEBUG = False

VERSION = "0.0.10"

"""
faces
=====

Provides fontfaces for the StyledTextCtrl for different platforms

"""

if wx.Platform == '__WXMSW__':
    faces = { 'times': 'Times New Roman',
              'mono' : 'Courier New',
              'helv' : 'Arial',
              'other': 'Comic Sans MS',
              'size' : 10,
              'size2': 8,
             }
elif wx.Platform == '__WXMAC__':
    faces = { 'times': 'Times New Roman',
              'mono' : 'Monaco',
              'helv' : 'Arial',
              'other': 'Comic Sans MS',
              'size' : 12,
              'size2': 10,
             }
else:
    faces = { 'times': 'Times',
              'mono' : 'Courier',
              'helv' : 'Helvetica',
              'other': 'new century schoolbook',
              'size' : 12,
              'size2': 10,
             }

"""
icontheme
=========

Provides the dict 'icons' with paths to the toolbar icons.

"""

# Toolbar items

icon_size = (36, 36)
icons = {"FileNew": LIBPREFIX + "icons/actions/filenew.png", \
         "FileOpen": LIBPREFIX + "icons/actions/fileopen.png", \
         "FileSave": LIBPREFIX + "icons/actions/filesave.png", \
         "FilePrint": LIBPREFIX + "icons/actions/fileprint.png", \
         "EditCut": LIBPREFIX + "icons/actions/edit-cut.png", \
         "EditCopy": LIBPREFIX + "icons/actions/edit-copy.png", \
         "EditCopyRes": LIBPREFIX + "icons/actions/edit-copy-results.png", \
         "EditPaste": LIBPREFIX + "icons/actions/edit-paste.png",
         "Undo": LIBPREFIX + "icons/actions/edit-undo.png",
         "Redo": LIBPREFIX + "icons/actions/edit-redo.png",
         "Find": LIBPREFIX + "icons/actions/edit-find.png",
         "FindReplace": LIBPREFIX + "icons/actions/edit-find-replace.png",}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# generated by wxGlade 0.6 on Sun May 25 23:31:23 2008

# Copyright 2008 Martin Manns
# Distributed under the terms of the GNU General Public License
# generated by wxGlade 0.6 on Mon Mar 17 23:22:49 2008

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""
_interfaces
===========

Provides:
---------
  1. CsvImport: Connects array datastructure with csv import module
  2. Clipboard: Clipboard access
"""

import csv

import wx

class CsvImport(object):
    """
    CSV file handling class
    
    Parameters:
    -----------
    filename: string
    \tFilename of CSV input file
    csvargs: dict
    \tArguments for csv import via the csv module
    
    Methods
    -------
    fill_target: Fills target with csv values.
    
    """
    def __init__(self, filename, csvargs):
        self.filename = filename
        self.csvargs = csvargs
        self.csvfile = None
    
    def _open_csv(self):
        """Opens the csv file named self.filename"""
        self.csvfile = open(self.filename, "r")
        return csv.reader(self.csvfile, **self.csvargs)

    def _close_csv(self):
        """Closes the csv file"""
        self.csvfile.close()
    
    def fill_target(self, target, key=(0, 0, 0)):
        """Fills target with csv values.
        
        The targets dimensions are assumed as defaulot limits
        unless specified otherwise.
        
        Parameters:
        -----------
        target: Mutable array or list with 2 dimensions
        \tIn this array, the csv values are stored
        
        key: 2-tuple, defaults to (0, 0)
        \tTop-left insertion position
        
        """
        
        error_log = []
        maxrows = len(target)
        top, left, table = key
        csv_it = self._open_csv()
        for row, values in enumerate(csv_it):
            if row != []:
                for col, ele in enumerate(values):
                    try:
                        target[row + top, col + left, table] = repr(ele)
                    except IndexError, err:
                        error_log.append([err, " at " + repr((col, ele))])
                if row > maxrows:
                    break
        self._close_csv()
        return error_log

# end of class CsvImport

class Clipboard(object):
    """Clipboard access
    
    Provides:
    ---------
    get_clipboard: Get clipboard content
    set_clipboard: Set clipboard content
    grid_paste: Inserts data into grid target
    
    """
    
    clipboard = wx.TheClipboard
    
    def _convert_clipboard(self, datastring=None, sep='\t'):
        """Converts data string to iterable.
        
        Parameters:
        -----------
        datastring: string, defaults to None
        \tThe data string to be converted. 
        \tself.get_clipboard() is called if set to None
        sep: string
        \tSeparator for columns in datastring
        
        """
        
        if datastring is None:
            datastring = self.get_clipboard()
        
        data_it = ((ele for ele in line.split(sep)) \
                            for line in datastring.splitlines())
        return data_it
        
    def get_clipboard(self):
        """Returns the clipboard text content"""
        
        textdata = wx.TextDataObject()
        if self.clipboard.Open():
            self.clipboard.GetData(textdata)
            self.clipboard.Close()
        else:
            wx.MessageBox("Can't open the clipboard", "Error")
        return textdata.GetText()
    
    def set_clipboard(self, data):
        """Writes data to the clipboard"""
        
        error_log = []
        
        textdata = wx.TextDataObject()
        try:
            textdata.SetText(data)
        except UnboundLocalError, err:
            error_log.append([err, "Error converting to string"])
        if self.clipboard.Open():
            self.clipboard.SetData(textdata)
            self.clipboard.Close()
        else:
            wx.MessageBox("Can't open the clipboard", "Error")
        return error_log
    
    def grid_paste(self, target, pos=(0, 0)):
        """Inserts data into target.
        
        The data in the clipboard has to be a whitespace separated
        2D matrix.
        
        Parameters:
        -----------
        target: Mutable array or list with 2 dimensions
        \tTarget array for data
        
        """
        error_log = []
        
        data_it = self._convert_clipboard()

        top, left, table = pos

        for row, values in enumerate(data_it):
            if row != []:
                for col, ele in enumerate(values):
                    if ele != '\b': # \b stands for omitted field
                        try:
                            target[row + top, col + left, table] = str(ele)
                        except IndexError, err:
                            error_log.append([err, " at " + repr((col, ele))])
        return error_log

# end of class Clipboard


def string_match(datastring, findstring, flags=None):
    """Returns True if findstring is found in datastring, otherwise False.
       Flags is a list of strings. Supported strings are:
       "MATCH_CASE": The case has to match for valid find
       "WHOLE_WORD": The word has to be surrounded by whitespace characters
                     if in the middle of the string
    """
    if flags is None:
        flags = []
    if "MATCH_CASE" not in flags:
        datastring = datastring.lower()
        findstring = findstring.lower()
    if findstring in datastring:
        if "WHOLE_WORD" in flags:
            findstringlength = len(findstring)
            try:
                if datastring.find(findstring) == 0 and \
                   datastring[findstringlength].isspace():
                    return 0
            except IndexError:
                if datastring == findstring: return 0
            try:
                if datastring.lstrip() == findstring and \
                   datastring[findstringlength + \
                   datastring.find(findstring)].isspace():
                    return datastring.find(findstring)
            except IndexError:
                if datastring.lstrip() == findstring: 
                    return datastring.find(findstring)
            for i, s in enumerate(datastring):
                try:
                    stringslice = datastring[i+1:i+1+findstringlength]
                    stringfound = (stringslice == findstring)
                except IndexError: 
                    stringfound = False
                if stringfound and s.isspace():
                    try:
                        if datastring[i+1+findstringlength].isspace():
                            return i + 1
                    except IndexError: pass
            try:
                if datastring.find(findstring) == len(datastring) - 1 and \
                   datastring[-findstringlength-1].isspace():
                    return len(datastring) - len(findstring) - 1
            except IndexError: pass
            return None
        else: # No "WHOLE_WORD"
            return datastring.find(findstring)
    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2008 Martin Manns
# Distributed under the terms of the GNU General Public License
# generated by wxGlade 0.6 on Mon Mar 17 23:22:49 2008

"""
pyspread
=======

Pyspread is a 3D spreadsheet application. 

Each cell accepts a Python expression and returns an accessible object. 
Python modules are usable from the spreadsheet table without external 
scripts.

Start from command line
pyspread.py

or from the Python shell:
>>> import pyspread
>>> pyspread.Pyspread()

Requires:
Python >= 2.4
Numpy >= 1.1.0
wxPython >= 2.8.7.1.

Examples require:
gmpy
rpy

--------------------------------------------------------------------
pyspread is distributed under the terms of the GNU General Public
License:

pyspread is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

pyspread is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------

"""

import bz2
import os
import sys
import csv
import cPickle as pickle

import numpy
import wx
import wx.grid
import wx.lib.printout as printout

import _widgets
from _choicebars import MainMenu, MainToolbar
from _dialogs import MacroDialog, CsvImportDialog, DimensionsEntryDialog, \
                     AboutDialog
from _interfaces import CsvImport, string_match
from config import DEBUG, LIBPREFIX, icon_size

# begin wxGlade: extracode
# end wxGlade


import distutils.sysconfig
try:
    testfile = open(LIBPREFIX + "icons/pyspread.png",'r')
    testfile.close()
except IOError:
    LIBPREFIX = distutils.sysconfig.get_python_lib() + '/pyspread/'


class MainWindow(wx.Frame):
    def __init__(self, *args, **kwds):
        try:
            dim = kwds.pop("dimensions")
        except:
            dim = (1000, 100, 1)
        
        self.wildcard = "Pyspread file (*.pys)|*.pys|"     \
                        "All files (*.*)|*.*"
                        
        kwds["style"] = wx.DEFAULT_FRAME_STYLE
        wx.Frame.__init__(self, *args, **kwds)
        
        # Menu Bar
        self.main_window_menubar = wx.MenuBar()
        self.main_menu = MainMenu(parent=self, menubar=self.main_window_menubar)
        self.SetMenuBar(self.main_window_menubar)
        
        # Status bar
        self.main_window_statusbar = self.CreateStatusBar(1, wx.ST_SIZEGRIP)
        
        # Tool Bar
        self.main_window_toolbar = MainToolbar(self, -1)
        self.SetToolBar(self.main_window_toolbar)
        
        # Main grid
        self.MainGrid = _widgets.MainGrid(self, -1, size=(1, 1), dim=dim)

        self.__set_properties()
        self.__do_layout()

        self.Bind(wx.EVT_FIND, self.OnFind)
        self.Bind(wx.EVT_FIND_NEXT, self.OnFind)
        self.Bind(wx.EVT_FIND_REPLACE, self.OnFind)
        self.Bind(wx.EVT_FIND_REPLACE_ALL, self.OnFind)
        self.Bind(wx.EVT_FIND_CLOSE, self.OnFindClose)
        wx.EVT_KEY_DOWN(self.MainGrid, self.OnKey)
        
        # Misc
        self.MainGrid.mainwindow = self
        self.MainGrid.deletion_imminent = False
           
    def __set_properties(self):
        self.SetTitle("pyspread")
        _icon = wx.EmptyIcon()
        _icon.CopyFromBitmap(wx.Bitmap(LIBPREFIX+'icons/pyspread.png', wx.BITMAP_TYPE_ANY))
        self.SetIcon(_icon)
        self.SetSize((1000, 700))
        self.main_window_statusbar.SetStatusWidths([-1])
        
        # statusbar fields
        main_window_statusbar_fields = [""]
        for i in range(len(main_window_statusbar_fields)):
            self.main_window_statusbar.SetStatusText(main_window_statusbar_fields[i], i)
        self.main_window_toolbar.SetToolBitmapSize(icon_size)
        self.main_window_toolbar.SetMargins((1, 1))
        self.main_window_toolbar.Realize()
        # end wxGlade
        self.MainGrid.create_rowcol()
        
    def __do_layout(self):
        # begin wxGlade: MainWindow.__do_layout
        sizer_1 = wx.BoxSizer(wx.VERTICAL)
        sizer_2 = wx.BoxSizer(wx.HORIZONTAL)
        sizer_3 = wx.BoxSizer(wx.VERTICAL)
        grid_sizer_2 = wx.FlexGridSizer(1, 1, 0, 0)
        sizer_3.Add(self.MainGrid.entry_line, 0, wx.EXPAND, 0)
        grid_sizer_2.Add(self.MainGrid, 0, wx.EXPAND, 0)
        grid_sizer_2.AddGrowableRow(0)
        grid_sizer_2.AddGrowableCol(0)
        sizer_3.Add(grid_sizer_2, 1, wx.EXPAND, 0)
        sizer_2.Add(sizer_3, 1, wx.EXPAND, 0)
        sizer_1.Add(sizer_2, 1, wx.EXPAND, 0)
        self.SetSizer(sizer_1)
        self.Layout()
        # end wxGlade
        
    def OnKey(self, event):
        ##print event.GetKeyCode()
        # If Ctrl+X is pressed...
        if event.ControlDown() and event.GetKeyCode() == 88:
            if DEBUG: print "Ctrl+X"
            # Call copy method
            self.OnCut()
        
        # If Ctrl+C is pressed...
        if not event.ShiftDown() and event.ControlDown() and event.GetKeyCode() == 67:
            if DEBUG: print "Ctrl+C"
            # Call copy method
            self.OnCopy()
        
        # If Shift+Ctrl+C is pressed...
        if event.ShiftDown() and event.ControlDown() and event.GetKeyCode() == 67:
            if DEBUG: print "Shift+Ctrl+C"
            # Call copy method
            self.OnCopyResult()
        
        # If Ctrl+V is pressed...
        if event.ControlDown() and event.GetKeyCode() == 86:
            if DEBUG: print "Ctrl+V"
            # Call paste method
            self.OnPaste()
            
        # If Supr is presed
        if event.GetKeyCode() == 127:
            if DEBUG: print "Supr"
            # Call delete method
            self.MainGrid.delete()
            
        # Skip other Key events
        if event.GetKeyCode():
            event.Skip()
        
    def OnFileNew(self, event): # wxGlade: MainWindow.<event_handler>
        dim_dialog = DimensionsEntryDialog(self)
        if dim_dialog.ShowModal() == wx.ID_OK:
            dim = dim_dialog.dimensions
            for i, d in enumerate(dim):
                if d <= 0: dim[i] = 1
            no_gridcells = reduce(lambda x, y: x*y, dim)
            if no_gridcells > 1E7:
                dlg = wx.MessageDialog(self, 'The grid has ' + \
                                       str(no_gridcells) + ' cells.\n' + \
                                       'You may run out of memory.\n' + \
                                       'Do you want to continue?',
                                       'Memory alert', 
                                       wx.OK | wx.CANCEL | wx.ICON_INFORMATION)
                res = dlg.ShowModal()
                if res != wx.ID_OK:
                    dim_dialog.Destroy()
                    event.Skip()
                    return 0
                dlg.Destroy()
            dimensions = tuple(dim)
            if DEBUG: print dimensions
            self.Destroy()
            self = MainWindow(None, dimensions=dimensions)
            self.Show()
        dim_dialog.Destroy()
        self.MainGrid.pysgrid.unredo.reset()
        event.Skip()

    def OnFileOpen(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnFileOpen'"
        dlg = wx.FileDialog(
            self, message="Choose a file", defaultDir=os.getcwd(), 
            defaultFile="", wildcard=self.wildcard, \
            style=wx.OPEN | wx.CHANGE_DIR)
        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPath()
            self.MainGrid.loadfile(path)
            self.MainGrid.OnCombo(event)
        event.Skip()

    def OnFileSave(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnFileSave'"
        dlg = wx.FileDialog(
            self, message="Save file as ...", defaultDir=os.getcwd(), 
            defaultFile="", wildcard=self.wildcard, style=wx.SAVE
            )
        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPath()
            print path
            outfile = bz2.BZ2File(path,"wb")
            pickle.dump(self.MainGrid.pysgrid.sgrid, outfile, protocol=2)
        dlg.Destroy()        
        event.Skip()

    def OnFilePrint(self, event): # wxGlade: MainWindow.<event_handler>
        ### UGLY CODE HERE, WE NEED A BRUSH-UP ###
        if DEBUG: print "Event handler `OnFilePrint'"
        selection = self.MainGrid.get_selection()
        rowslice, colslice = self.MainGrid.get_selected_rows_cols(selection)
        data = self.MainGrid.getselectiondata(self.MainGrid.pysgrid, \
                                              rowslice, colslice)
        prt = printout.PrintTable(self.MainGrid)
        prt.data = data
        prt.Print()
        event.Skip()

    def OnExit(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnExit'"
        sys.exit()
        event.Skip()

    def OnCut(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnCut'"
        self.MainGrid.cut()
        self.MainGrid.pysgrid.unredo.mark()
        event.Skip()

    def OnCopy(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnCopy'"
        self.MainGrid.copy(source=self.MainGrid.pysgrid.sgrid)
        self.MainGrid.pysgrid.unredo.mark()
        event.Skip()

    def OnCopyResult(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnCopyResult'"
        self.MainGrid.copy(source=self.MainGrid.pysgrid)
        self.MainGrid.pysgrid.unredo.mark()
        event.Skip()

    def OnPaste(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnPaste'"
        self.MainGrid.paste()
        self.MainGrid.pysgrid.unredo.mark()
        event.Skip()

    def OnAbout(self, event): # wxGlade: MainWindow.<event_handler>
        about_dialog = AboutDialog(self)
        about_dialog.ShowModal()
        about_dialog.Destroy()
        event.Skip()

    def OnInsertRows(self, event): # wxGlade: MainWindow.<event_handler>
        """Insert the maximum of 1 and the number of selected rows""" 
        if DEBUG: print "Event handler `OnInsertRows'"
        self.MainGrid.insert_rows()
        event.Skip()

    def OnInsertColumns(self, event): # wxGlade: MainWindow.<event_handler>
        """Inserts the maximum of 1 and the number of selected columns """
        if DEBUG: print "Event handler `OnInsertColumns'"
        self.MainGrid.insert_cols()
        event.Skip()

    def OnInsertTable(self, event): # wxGlade: MainWindow.<event_handler>
        """Insert one table into MainGrid and pysgrid """ 
        if DEBUG: print "Event handler `OnInsertTable'"
        self.MainGrid.insert_tables()
        event.GetString = lambda x=0: str(self.MainGrid.current_table)
        self.MainGrid.OnCombo(event)
        event.Skip()

    def OnDeleteRows(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnDeleteRows'"
        self.MainGrid.delete_rows()
        event.Skip()
        
    def OnDeleteColumns(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnDeleteColumns'"
        self.MainGrid.delete_cols()
        event.Skip()

    def OnDeleteTable(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnDeleteTable'"
        self.MainGrid.delete_tables()
        event.GetString = lambda x=0: \
              str(max(0, self.MainGrid.current_table, \
                         self.MainGrid.pysgrid.sgrid.shape[2]-1))
        self.MainGrid.OnCombo(event)
        event.Skip()

    def OnCSVImport(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnCSVImport'"
        csvwildcard = " CSV file|*.*|Tab-delimited text file|*.*"
        path = None
        # File choice
        filedlg = wx.FileDialog(
            self, message="Import a csv-file", defaultDir=os.getcwd(), 
            defaultFile="", wildcard=csvwildcard, style=wx.OPEN | wx.CHANGE_DIR
            )
        if filedlg.ShowModal() == wx.ID_OK:
            path = filedlg.GetPath()
            filedlg.Destroy()
        
        # CSV import option choice
        if path is None:
            return 0 
        filterdlg = CsvImportDialog(self, csvfilepath=path)
        if filedlg.GetFilterIndex() == 0:
            if filterdlg.ShowModal() == wx.ID_OK:
                csvargs = filterdlg.get_csvargs()
            else:
                filterdlg.Destroy()
                return 0                
        elif filedlg.GetFilterIndex() == 1:
            csvargs = {'dialect': 'excel-tab', 'quoting': csv.QUOTE_NONNUMERIC}
        
        # The actual data import
        csv_interface = CsvImport(path, csvargs)
        topleftcell = tuple(list(self.MainGrid.get_currentcell()) + \
                            [self.MainGrid.current_table])
        csv_interface.fill_target(self.MainGrid.pysgrid, key=topleftcell)
        self.MainGrid.update_grid(funceval=True, gridslice=None)
        event.Skip()

    def OnMacroList(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnMacroList'"
        #print self.MainGrid.pysgrid.macros
        dlg = MacroDialog(None, -1, "", macros = self.MainGrid.pysgrid.macros)
        if dlg.ShowModal() == wx.ID_OK:
            # Insert function string into current cell
            targetcell = self.MainGrid.get_currentcell()
            macrostring = dlg.GetMacroString()
            try:
                self.MainGrid.entry_line.SetValue(macrostring)
            except TypeError: 
                self.MainGrid.entry_line.SetValue("")
            self.MainGrid.pysgrid[targetcell] = macrostring
            self.MainGrid.update_grid(funceval=True)
        self.MainGrid.pysgrid.macros = dlg.macros
        self.MainGrid.pysgrid.set_global_macros(self.MainGrid.pysgrid.macros)
        dlg.Destroy()
        self.MainGrid.update_grid()
        event.Skip()

    def OnMacroListLoad(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnMacroListLoad'"
        macrowildcard = " Macro file|*.*"
        # File choice
        filedlg = wx.FileDialog(
            self, message="Load a Macro-file", defaultDir=os.getcwd(), 
            defaultFile="", wildcard=macrowildcard, \
            style=wx.OPEN | wx.CHANGE_DIR)
        if filedlg.ShowModal() == wx.ID_OK:
            path = filedlg.GetPath()
            filedlg.Destroy()
        macrocodes = {}
        infile = bz2.BZ2File(path, "r")
        macrocodes = pickle.load(infile)
        infile.close()
        #print macrocodes
        for macroname in macrocodes:
            self.MainGrid.pysgrid.macros.AddToMacroDict(macrocodes[macroname])
        self.MainGrid.pysgrid.set_global_macros()
        event.Skip()

    def OnMacroListSave(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnMacroListSave'"
        macrowildcard = " Macro file|*.*"
        # File choice
        filedlg = wx.FileDialog(
            self, message="Save a Macro-file", defaultDir=os.getcwd(), 
            defaultFile="", wildcard=macrowildcard, \
            style=wx.OPEN | wx.CHANGE_DIR)
        if filedlg.ShowModal() == wx.ID_OK:
            path = filedlg.GetPath()
            filedlg.Destroy()
        macros = self.MainGrid.pysgrid.macros
        macrocodes = dict((m, macros[m].func_dict['macrocode']) for m in macros)
        outfile = bz2.BZ2File(path,"w")
        pickle.dump(macrocodes, outfile, protocol=2)
        outfile.close()
        event.Skip()
    
    def OnFind(self, event): # wxGlade: MainWindow.<event_handler>
        ### Find functionality should be in interfaces
        if DEBUG: print "Event handler `OnFind'"
        wx_map = { wx.wxEVT_COMMAND_FIND : "FIND",
                   wx.wxEVT_COMMAND_FIND_NEXT : "FIND_NEXT",
                   wx.wxEVT_COMMAND_FIND_REPLACE : "REPLACE",
                   wx.wxEVT_COMMAND_FIND_REPLACE_ALL : "REPLACE_ALL" }
        wx_flags = { 0: ["UP", ],
                     1: ["DOWN"], 
                     2: ["UP", "WHOLE_WORD"], 
                     3: ["DOWN", "WHOLE_WORD"], 
                     4: ["UP", "MATCH_CASE"], 
                     5: ["DOWN", "MATCH_CASE"], 
                     6: ["UP", "WHOLE_WORD", "MATCH_CASE"], 
                     7: ["DOWN", "WHOLE_WORD", "MATCH_CASE"] }
        et = event.GetEventType()
        ef = event.GetFlags()        
        try:
            event_type = wx_map[et]
            event_flags = wx_flags[ef]
        except KeyError:
            if DEBUG: print "Unknown event type or flag:", et, ef
            return 0            
        event_find_string = event.GetFindString()
        event_replace_string = event.GetReplaceString()
        
        findpos = self.MainGrid.pysgrid.findnextmatch(self.MainGrid.key, \
                                            event_find_string, event_flags)
        if event_type in ["FIND", "FIND_NEXT", "REPLACE"]:
            if findpos is not None:
                self.MainGrid.selectnewcell(findpos, event)
                self.main_window_statusbar.SetStatusText("Found '" + \
                                 event_find_string +  "' in cell " + \
                                 str(list(findpos)) + ".", 0)
            else:
                self.main_window_statusbar.SetStatusText("'" + \
                                 event_find_string + "' not found.", 0)
        if event_type in ["REPLACE", "REPLACE_ALL"]:
            noreplaced = 0
            while findpos is not None:
                noreplaced += 1
                cellstring = self.MainGrid.pysgrid.sgrid[findpos]
                string_position = string_match(cellstring, \
                                    event_find_string, event_flags)
                newstring = cellstring[:string_position]
                newstring += cellstring[string_position:].replace(\
                                       event_find_string, \
                                       event_replace_string, 1)
                self.MainGrid.pysgrid.sgrid[findpos] = newstring
                if event_type == "REPLACE":
                    self.main_window_statusbar.SetStatusText("'" + 
                                 cellstring + "' replaced by '" + \
                                 newstring + "'.", 0)
                    break
                elif event_type == "REPLACE_ALL":
                    findpos = self.MainGrid.pysgrid.findnextmatch( \
                                 findpos, event_find_string, event_flags)
                else: 
                    raise ValueError, "Event type " + event_type + " unknown."
            self.MainGrid.update_grid()
            #self.__fill_cells()
            if event_type == "REPLACE_ALL":
                self.main_window_statusbar.SetStatusText(str(noreplaced) + \
                                 " occurrences of '" + event_find_string + \
                                 "' replaced by '" +  event_replace_string + \
                                 "'.", 0)
            else:
                self.MainGrid.entry_line.SetValue(newstring)
        event.Skip()

    def OnFindClose(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnFindClose'"
        event.GetDialog().Destroy()
        self.main_window_statusbar.SetStatusText("", 0)
        event.Skip()


    def OnUndo(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnUndo'"
        self.MainGrid.undo()
        event.Skip()

    def OnRedo(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnRedo'"
        self.MainGrid.redo()
        event.Skip()

    def OnShowFind(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnShowFind'"
        data = wx.FindReplaceData()
        dlg = wx.FindReplaceDialog(self, data, "Find")
        dlg.data = data  # save a reference to it...
        dlg.Show(True)        
        event.Skip()

    def OnShowFindReplace(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnShowFindReplace'"
        data = wx.FindReplaceData()
        dlg = wx.FindReplaceDialog(self, data, "Find & Replace", \
                                   wx.FR_REPLACEDIALOG)
        dlg.data = data  # save a reference to it...
        dlg.Show(True)        
        event.Skip()

# end of class MainWindow

def main(filename=None, dimensions=(1000, 100, 3)):
    """
    Starts pyspread with file 'filename' or with empty grid
    
    Parameters
    ----------
    filename: String defaults to None
    \tfilename of the pyspread file that is initially loaded
    dimensions: 3-tuple
    \tDimensions of empty grid (i.e. if filename is None) (rows, cols, tables)
    
    """
    app = wx.PySimpleApp(0)
    wx.InitAllImageHandlers()
    main_window = MainWindow(None, dimensions=dimensions)
    if filename is not None:
        main_window.MainGrid.loadfile(filename)
    app.SetTopWindow(main_window)
    main_window.Show()
    app.MainLoop()

if __name__ == "__main__":
    #cProfile.run('main()')
    try:
        filename = sys.argv[1]
    except IndexError:
        filename = None    
    main(filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# generated by wxGlade 0.6 on Sun May 25 23:31:23 2008

# Copyright 2008 Martin Manns
# Distributed under the terms of the GNU General Public License
# generated by wxGlade 0.6 on Mon Mar 17 23:22:49 2008

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""
_widgets
========

Provides:
---------
  1. IndexGrid: Grid that supports indexing and __len__
  2. MainGrid: Main grid
  2. CollapsiblePane: Collapsible pane with basic toggle mechanism
  3. MacroEditPanel: Collapsible label, parameter entry area and editor
  4. SortedListCtrl: ListCtrl with items sorted in first column

"""

import bz2
import keyword
import cPickle as pickle
from itertools import izip

import wx
import wx.grid
import wx.stc  as  stc
import wx.lib.mixins.listctrl  as  listmix
import numpy

from _datastructures import PyspreadGrid
from _choicebars import ContextMenu
from _interfaces import Clipboard
from config import faces, icon_size, DEBUG


class IndexGrid(wx.grid.Grid):
    """Grid that supports indexing and __len__"""
    def __getitem__(self, key):
        if type(key) is tuple:
            row, col = key[:2]
            if row < 0:
                row += self.GetNumberRows()
            if row >= 0:
                return self.GetCellValue(row, col)
        raise KeyError("key must be a tuple of length 2")

    def __setitem__(self, key, value):
        if type(key) is tuple:
            row, col = key[:2]
            if row < 0:
                row += self.GetNumberRows()
            if row >= 0:
                return self.SetCellValue(row, col, value)
        raise KeyError("key must be a tuple of length 2")

    def __len__(self):
        return self.GetNumberRows()

# end of class IndexGrid


class MainGrid(IndexGrid):
    """Main grid"""
    def __init__(self, parent, id=wx.NewId(), pos=(0, 0), size=(1, 1), style=0, \
                 mainwindow=None, datagrid=None, dim=(1000, 100, 3)):
        # Sub widgets
        self.parent = parent
        self.mainwindow = mainwindow
        self.entry_line = wx.TextCtrl(parent, -1, "")        
        self.pysgrid = PyspreadGrid(dim)
        self.__add_combo(parent)
        self.contextmenu = ContextMenu(parent=parent)
        self.clipboard = Clipboard()
        
        self.copy_selection = [] # Cells from last copy operation 
        self.current_table = 0
        self.key = (0, 0, 0)
        self.selection_imminent = False
        
        wx.grid.Grid.__init__(self, parent, id, pos, size, style)
        
        self.Bind(wx.grid.EVT_GRID_CELL_RIGHT_CLICK, self.OnContextMenu)
        
        self.Bind(wx.grid.EVT_GRID_CMD_CELL_CHANGE, self.OnCellChanged)
        self.Bind(wx.grid.EVT_GRID_CMD_EDITOR_SHOWN, self.OnCellEditorShown)
        self.Bind(wx.grid.EVT_GRID_CMD_SELECT_CELL, self.OnCellSelected)
        self.Bind(wx.EVT_SCROLLWIN, self.OnScroll)
        self.entry_line.Bind(wx.EVT_TEXT, self.OnText)
        self.entry_line.Bind(wx.EVT_CHAR, self.EvtChar)
        

    def __add_combo(self, parent):
        """Choicebox for the Z choice of the worksheet"""
        cbID = wx.NewId()
        self.cbox_Z = wx.ComboBox(parent.main_window_toolbar, cbID, "0", \
                      choices=[str(dim) for dim in xrange(self.pysgrid.sgrid.shape[2])],
                      size=(icon_size[0]*2, icon_size[1]-4), style=wx.CB_DROPDOWN)
        parent.main_window_toolbar.AddControl(self.cbox_Z)
            
        parent.Bind(wx.EVT_TEXT, self.OnCombo, id=cbID)
        parent.Bind(wx.EVT_COMBOBOX, self.OnCombo, id=cbID)
        parent.main_window_toolbar.Realize()
        
    def create_rowcol(self):
        """Creates a new, empty grid"""
        
        operation = (self.create_rowcol, [])
        undo_operation = (self.create_rowcol, [])
        self.pysgrid.unredo.append(undo_operation, operation)
        
        try:
            self.CreateGrid(*self.pysgrid.sgrid.shape[:2])
        except:
            pass
        for i in xrange(self.pysgrid.sgrid.shape[0]):
            self.SetRowLabelValue(i, str(i))
        for i in xrange(self.pysgrid.sgrid.shape[1]):
            self.SetColLabelValue(i, str(i))

    def get_selected_rows_cols(self, selection = None):
        """ Get the slices of selected rows and cols, None if no selection """
        if selection is None: selection = self.get_selection()
        try: rowslice = slice(min(c[0] for c in selection), \
                              max(c[0] for c in selection) + 1)
        except IndexError: rowslice = None
        try: colslice = slice(min(c[1] for c in selection), \
                              max(c[1] for c in selection) + 1)
        except IndexError: colslice = None
        return rowslice, colslice
    
    def get_selection(self):
        """ Returns an index list of all cells that are selected in the grid.
        All selection types are considered equal. If no cells are selected,
        the current cell is returned."""
                
        # GetSelectedCells: individual cells selected by ctrl-clicking
        # GetSelectedRows: rows selected by clicking on the labels
        # GetSelectedCols: cols selected by clicking on the labels
        # GetSelectionBlockTopLeft
        # GetSelectionBlockBottomRight: For blocks of cells selected by dragging 
        # across the grid cells.
              
        dimx, dimy = self.pysgrid.sgrid.shape[:2]

        selected_rows = self.GetSelectedRows()
        selected_cols = self.GetSelectedCols()
        
        selection = []
        selection += self.GetSelectedCells()
        selection += list((row, y) for row in selected_rows for y in xrange(dimy))
        selection += list((x, col) for col in selected_cols for x in xrange(dimx))
        
        selectionblock = izip(self.GetSelectionBlockTopLeft(), \
                              self.GetSelectionBlockBottomRight())
        for tl, br in selectionblock:
            selection += [(x, y) for x in xrange(tl[0], br[0]+1) 
                                 for y in xrange(tl[1], br[1]+1)]
        
        if selection == []:
            selection = [(self.get_currentcell())]
        selection = sorted(list(set(selection)))
        return selection
                    
    def get_currentcell(self):
        # Get cursor position
        row = self.GetGridCursorRow()
        col = self.GetGridCursorCol()
        return row, col
    
    def get_visiblecell_slice(self):
        topleft_x = self.YToRow(self.GetViewStart()[1] * self.ScrollLineX)
        topleft_y = self.XToCol(self.GetViewStart()[0] * self.ScrollLineY)
        topleft = (topleft_x, topleft_y)
        #print topleft
        row, col = topleft_x + 1, topleft_y + 1 # Ensures visibility
        #print self.IsVisible(*topleft)
        while self.IsVisible(row, topleft[1], wholeCellVisible=False):
            row += 1
        while self.IsVisible(topleft[0], col, wholeCellVisible=False):
            col += 1
        lowerright = (row, col) # This cell is *not* visible
        return (slice(topleft[0], lowerright[0]), \
                slice(topleft[1], lowerright[1]), 
                slice(self.current_table, self.current_table+1))
                
    def __fill_cells(self, gridslice=None):
        """Fills grid"""
        for i in self.pysgrid.get_function_cell_indices(gridslice):
            if i[2] == self.current_table:
                self.SetCellValue(i[0], i[1], str(self.pysgrid[i]))
        current_cell_index = (self.GridCursorRow, \
                              self.GridCursorCol, \
                              self.current_table)
        if current_cell_index == (-1, -1, self.current_table):
            current_cell_index = (0, 0, self.current_table)
        self.SetCellValue(current_cell_index[0], \
                          current_cell_index[1], \
                          str(self.pysgrid[current_cell_index]))

    def update_grid(self, funceval=False, funceval_gridslice=None, gridslice="None"):
        """Updates the main grid.
        
        Updates the function grid slice funceval_gridslice if funceval==True.
        
        """
        if funceval:
            self.pysgrid.eval_function_grid(gridslice=funceval_gridslice)
        if gridslice == "None":
            gridslice = self.get_visiblecell_slice()
        self.__fill_cells(gridslice=gridslice)

    def selectnewcell(self, key, event):
        """ Selects the cell with position key.
        
        Parameters:
        -----------
        key: 3-tuple
        \tPosition in grid(x,y,z)
        """
        if key[2] != self.current_table:
            self.cbox_Z.SetSelection(key[2])
            event.GetString = lambda x=0: str(key[2])
            self.OnCombo(event)
            self.SetFocus()
        self.MakeCellVisible(*key[:2])
        self.SetGridCursor(*key[:2])

    def OnScroll(self, event):
        self.update_grid()
        self.Update()
        event.Skip()

    def OnCellEditorShown(self, event):
        if DEBUG: print "Event handler `OnCellEditorShown'"
        self.key = (event.Row, event.Col, self.current_table)
        self._set_cell_to_expression(self.key)
        event.Skip()

    def OnCellChanged(self, event):
        if DEBUG: print "Event handler `OnCellChanged'"
        self.key = (event.Row, event.Col, self.current_table)
        currstr = self[self.key[:2]]
        try:
            self.CreateGrid(*self.pysgrid.sgrid.shape[:2])
        except AssertionError:
            pass
        self.pysgrid[self.key] = currstr
        funcevalslice = tuple([slice(self.key[i], self.key[i]+1) \
                                              for i in xrange(3)])
        self.update_grid(funceval=True, funceval_gridslice=funcevalslice)
        self.pysgrid.unredo.mark()
        event.Skip()

    def OnText(self, event): # wxGlade: MainWindow.<event_handler>
        if DEBUG: print "Event handler `OnText'"
        if self.selection_imminent:
            self.selection_imminent = False
        else:
            currstr = event.GetString()
            self[self.key[:2]] = event.GetString()
            if currstr == '':
                self.pysgrid.fgrid[self.key] = 0
            self.pysgrid[self.key] = currstr
        try:
            currcellslice = (slice(self.key[0], self.key[0]+1), \
                             slice(self.key[1], self.key[1]+1))
            self.update_grid(funceval=True, \
                             funceval_gridslice=currcellslice, \
                             gridslice=currcellslice)
        except IndexError:
            pass
        event.Skip()
        
    def EvtChar(self, event):
        if event.GetKeyCode() == 13:
            self.update_grid()
            self.SetFocus()
        event.Skip()

    def OnCellSelected(self, event):
        if DEBUG: print "Event handler `OnCellSelected'"
        self.selection_imminent = True
        self.key = (event.Row, event.Col, self.current_table)
        try: 
            currstr = self.pysgrid.sgrid[self.key]
        except IndexError: 
            currstr = ""
        try: 
            self.entry_line.SetValue(currstr)
        except TypeError: 
            self.entry_line.SetValue("")
        self.update_grid()
        event.Skip()

    def OnContextMenu(self, event):
        if DEBUG: print "Event handler `OnContextMenu'"
        self.PopupMenu(self.contextmenu)
        #contextmenu.Destroy()
        event.Skip()
    
    def OnCombo(self, event):
        try:
            newtable = int(event.GetString())
        except ValueError:
            newtable = 0
        if newtable in xrange(self.pysgrid.sgrid.shape[2]):
            # Update the whole grid including the empty cells
            self.current_table = newtable
            self.ClearGrid()
            self.update_grid()
            # Now update the EntryLine to the current string
            self.pos = tuple(list(self.get_currentcell())+[self.current_table])
            currstr = self.pysgrid.sgrid[self.pos]
            if self.selection_imminent:
                event.Skip()
                return 0
            self.selection_imminent = True
            try:
                self.entry_line.SetValue(currstr)
            except TypeError:
                self.entry_line.SetValue("")
            self.selection_imminent = False
        event.Skip()

    def loadfile(self, filename):
        """Load file named fileneme into spreadsheet"""
        try:
            infile = bz2.BZ2File(filename, "rb")
        except IOError:
            return 0
        self.pysgrid.sgrid = pickle.load(infile)
        infile.close()
        self.DeleteRows(0, self.pysgrid.fgrid.shape[0])
        self.DeleteCols(0, self.pysgrid.fgrid.shape[1])
        self.AppendRows(self.pysgrid.sgrid.shape[0])
        self.AppendCols(self.pysgrid.sgrid.shape[1])
        self.cbox_Z.Clear()
        self.cbox_Z.AppendItems([str(i) for i in xrange(self.pysgrid.sgrid.shape[2])])
        self.pysgrid.fgrid = numpy.zeros(self.pysgrid.sgrid.shape, dtype="O")
        self.update_grid(funceval=True, funceval_gridslice=None, gridslice=None)
        self.create_rowcol()
        self.pysgrid.unredo.reset()

    def cut(self):
        """ Cuts TextCtrlSelection if present 
            else cuts Grid cells
            Source can be sgrid, fgrid or the displayed wxGrid """
        self.copy(source=self.pysgrid.sgrid)
        focus = self.parent.FindFocus()
        if isinstance(focus, wx.TextCtrl):
            self.selection_replace(focus, "")
        else:
            self.delete()

    def getselectiondata(self, source, rowslice, colslice, selection=None):
        """2D source data array that matches the current selection"""
        try:
            data = source[rowslice, colslice, self.current_table].copy()
        except AttributeError:
            data = source[rowslice, colslice, self.current_table]
        try:
            len(data)
        except TypeError:
            return data
        for row in xrange(len(data)):
            for col in xrange(len(data[row])):
                if selection is not None and \
                   (row + rowslice.start, col + colslice.start) not in selection:
                    data[row, col] = '\b'  # \b stands for omitted field
        return data

    def _set_cell_to_expression(self, key):
        """Update MainGrid cell from result to python expression"""
        try:
            self[key[:2]] = self.pysgrid.sgrid[key]
        except TypeError: 
            self[key[:2]] = ''

    def copy(self, source=None):
        """Copies TextCtrlSelection if present else copies Grid cells
        
        Parameters
        ----------
        source: sgrid or fgrid
        
        """
        if source is None: source = self.pysgrid.sgrid
        focus = self.parent.FindFocus()
        if isinstance(focus, wx.TextCtrl):
            sel = focus.GetStringSelection()
            if sel != "":
                clipboard_data = sel
                self.copy_selection = []
            else: return 0
        else:            
            selection = self.get_selection()
            #print selection
            rowslice, colslice = self.get_selected_rows_cols(selection)
            data = self.getselectiondata(source, rowslice, colslice, selection)
            #print data
            try:
                clipboard_data = "\n".join("\t".join(str(d) for d in datarow) for datarow in data)
            except TypeError:
                clipboard_data = str(data)
            self.copy_selection = [(cell[0]-rowslice.start, cell[1]-colslice.start) for cell in selection]
        self.clipboard.set_clipboard(clipboard_data)
        
    def paste(self):
        """Pastes into TextCtrl if active else pastes to grid"""
        
        focus = self.parent.FindFocus()
        if isinstance(focus, wx.TextCtrl):
            data = self.clipboard.get_clipboard()
            self.selection_replace(focus, data)
        else: # We got a grid selection
            pastepos = (self.GetGridCursorRow(), self.GetGridCursorCol(), self.current_table)
            self.clipboard.grid_paste(self.pysgrid, pos=pastepos)
        self.update_grid(funceval=True)
    
    def undo(self):
        self.pysgrid.unredo.undo()
        self.ClearGrid()
        self.update_grid(funceval=True)
        self.create_rowcol()
        
    def redo(self):
        self.pysgrid.unredo.redo()
        self.ClearGrid()
        self.update_grid(funceval=True)
        self.create_rowcol()

    def selection_replace(self, editor, data):
        """ Replaces a selection in a TextCtrl with inserted data"""
        ##*** This should be moved into a custom TextCtrl class ***
        inspoint = int(editor.InsertionPoint)
        sel_begin, sel_end = editor.GetSelection()
        #print sel_begin, sel_end
        if sel_begin != sel_end and inspoint > sel_begin:
            inspoint = inspoint - min(abs(sel_end - sel_begin), abs(inspoint - sel_begin))
        oldval = editor.GetValue()[:sel_begin] + editor.GetValue()[sel_end:]
        newval = oldval[:inspoint] + data + oldval[inspoint:]
        editor.SetValue(newval)
        editor.SetInsertionPoint(inspoint + len(data))
        editor.SetSelection(inspoint, inspoint + len(data))
        
    def delete(self, selection=None):
        if selection is None:
            selection = self.get_selection()
        for cell in selection:
            self.pysgrid.fgrid[cell[0], cell[1], self.current_table] = 0
            self.pysgrid[cell[0], cell[1], self.current_table] = ''
            self.SetCellValue(cell[0], cell[1], '')
        self.pysgrid.unredo.mark()

    def insert_rows(self):
        # Current insertion position and selected rows
        selectedrows = set(c[0] for c in self.get_selection())
        current_row = min(selectedrows)
        no_selected_rows = max(1, len(selectedrows))
        # Insert rows into MainGrid and pysgrid
        operation = (self.InsertRows, [current_row, no_selected_rows, False])
        undo_operation = (self.DeleteRows, [current_row, no_selected_rows, False])
        self.pysgrid.unredo.append(undo_operation, operation)
        self.InsertRows(current_row, no_selected_rows, updateLabels=False)
        
        self.pysgrid.insert(insertionpoint=[current_row, None, None], notoinsert=no_selected_rows)
        self.create_rowcol()
        self.pysgrid.unredo.mark()
    
    def insert_cols(self):
        # Current insertion position and selected columns
        selectedcols = set(c[1] for c in self.get_selection())
        current_col = min(selectedcols)
        no_selected_cols = max(1, len(selectedcols))
        # Insert columns into MainGrid and pysgrid
        operation = (self.InsertCols, [current_col, no_selected_cols, False])
        undo_operation = (self.DeleteCols, [current_col, no_selected_cols, False])
        self.pysgrid.unredo.append(undo_operation, operation)
        self.InsertCols(current_col, no_selected_cols, updateLabels=False)
        
        self.pysgrid.insert(insertionpoint=[None, current_col, None], notoinsert=no_selected_cols)
        self.create_rowcol()
        self.pysgrid.unredo.mark()
    
    def insert_tables(self):
        current_table = self.current_table
        no_selected_tables = 1
        operation = (self.cbox_Z.Append, [str(self.pysgrid.sgrid.shape[2])])
        undo_operation = (self.cbox_Z.Delete, [self.pysgrid.sgrid.shape[2]-1])
        self.pysgrid.unredo.append(undo_operation, operation)
        self.cbox_Z.Append(str(self.pysgrid.sgrid.shape[2]))
        self.pysgrid.insert(insertionpoint=[None, None, current_table], notoinsert=no_selected_tables)
        self.pysgrid.unredo.mark()

    def delete_rows(self):
        if self.pysgrid.sgrid.shape[0] > 0:
            # Current deletion position and selected rows
            selectedrows = set(c[0] for c in self.get_selection())
            current_row = min(selectedrows)
            no_selected_rows = min(max(1, len(selectedrows)), self.pysgrid.sgrid.shape[0] - 1)
            # Delete rows from MainGrid and pysgrid
            operation = (self.DeleteRows, [current_row, no_selected_rows, False])
            undo_operation = (self.InsertRows, [current_row, no_selected_rows, False])
            self.pysgrid.unredo.append(undo_operation, operation)
            self.DeleteRows(current_row, no_selected_rows, updateLabels=False)
            self.pysgrid.remove(removalpoint=[current_row, None, None], notoremove=no_selected_rows)
            self.create_rowcol()
        self.pysgrid.unredo.mark()
    
    def delete_cols(self):
        if self.pysgrid.sgrid.shape[1] > 0:
            # Current insertion position and selected columns
            selectedcols = set(c[1] for c in self.get_selection())
            current_col = min(selectedcols)
            no_selected_cols = min(max(1, len(selectedcols)), self.pysgrid.sgrid.shape[1] - 1)
            # Delete columns from MainGrid and pysgrid
            operation = (self.DeleteCols, [current_col, no_selected_cols, False])
            undo_operation = (self.InsertCols, [current_col, no_selected_cols, False])
            self.pysgrid.unredo.append(undo_operation, operation)
            self.DeleteCols(current_col, no_selected_cols, updateLabels=False)
            self.pysgrid.remove(removalpoint=[None, current_col, None], notoremove=no_selected_cols)
            self.create_rowcol()
        self.pysgrid.unredo.mark()
    
    def delete_tables(self):
        if self.pysgrid.sgrid.shape[2] > 1:
            current_table = self.current_table
            no_selected_tables = 1
            operation = (self.cbox_Z.Delete, [self.pysgrid.sgrid.shape[2] - 1])
            undo_operation = (self.cbox_Z.Append, [str(self.pysgrid.sgrid.shape[2])])
            self.pysgrid.unredo.append(undo_operation, operation)
            self.cbox_Z.Delete(self.pysgrid.sgrid.shape[2] - 1)
            self.pysgrid.remove(removalpoint=[None, None, current_table], notoremove=no_selected_tables)
        self.pysgrid.unredo.mark()

# end of class MainGrid

class CollapsiblePane(wx.CollapsiblePane):
    """Collapsible pane with basic toggle mechanism
    
    Parameters:
    -----------
    panename: string
    \tLabel for the collapsible pane
    
    """
    
    def __init__(self, *args, **kwds):
        self.label_show = "Show "
        self.label_hide = "Hide "
        panename = kwds.pop('panename')
        self.__dict__['label'] = panename
        wx.CollapsiblePane.__init__(self, *args, **kwds)
        self.SetLabel(self.label_show + panename)
        self.Bind(wx.EVT_COLLAPSIBLEPANE_CHANGED, self.restore_pane, self)
        
    def OnToggle(self, event):
        """"Toggle event handler"""
        self.Collapse(self.IsExpanded())
        self.restore_pane()
        
    def restore_pane(self, event=None):
        """Restores the layout of the content and changes teh labels"""
        self.GetParent().Layout()
        # and also change the labels
        if self.IsExpanded():
            self.SetLabel(self.label_hide + self.__dict__['label'])
        else:
            self.SetLabel(self.label_show + self.__dict__['label'])

# end of class CollapsiblePane

class MacroEditPanel(wx.Panel):
    """Collapsible label, parameter entry area and editor"""
    def __init__(self, parent,  *args, **kwds):
        wx.Panel.__init__(self, parent, -1)
        self.maxvar = 100 # The menu will fail for functions with more variables
        sizer = wx.FlexGridSizer(cols=1, rows=3, hgap=5, vgap=5)
        #print sizer.GetColWidths()
        sizer.AddGrowableCol(0)
        sizer.AddGrowableRow(2)
        self.SetSizer(sizer)
        # Add the three panes on the right hand side
        self.pane_docstring = CollapsiblePane(self, label='', panename = 'docstring', style=wx.CP_DEFAULT_STYLE|wx.CP_NO_TLW_RESIZE)
        self.make_docstring_content(self.pane_docstring.GetPane())
        sizer.Add(self.pane_docstring, 0, wx.RIGHT|wx.LEFT|wx.EXPAND, 25)
        self.pane_docstring.OnToggle(None)
        
        self.pane_macroform = CollapsiblePane(self, label='', panename = 'macro form', style=wx.CP_DEFAULT_STYLE|wx.CP_NO_TLW_RESIZE)
        self.make_macroform_content(self.pane_macroform.GetPane())
        sizer.Add(self.pane_macroform, 0, wx.RIGHT|wx.LEFT|wx.EXPAND, 25)
        self.pane_macroform.OnToggle(None)
        
        self.pane_code = CollapsiblePane(self, label='', panename = 'code', style=wx.CP_DEFAULT_STYLE|wx.CP_NO_TLW_RESIZE)
        self.make_code_content(self.pane_code.GetPane())
        sizer.Add(self.pane_code, 0, wx.RIGHT|wx.LEFT|wx.EXPAND, 25)
        
    def make_docstring_content(self, pane):
        self.docstringTextCtrl = wx.TextCtrl(pane, -1, '', style = wx.NO_BORDER|wx.TE_MULTILINE|wx.TE_READONLY)
        #self.docstringTextCtrl.SetEditable(False)
        sizer = wx.FlexGridSizer(cols=1, rows=1, hgap=5, vgap=5)
        sizer.AddGrowableCol(0)
        sizer.AddGrowableRow(0)
        sizer.Add(self.docstringTextCtrl, 0, wx.EXPAND|wx.FIXED_MINSIZE, 0) 
        pane.SetSizer(sizer)
    
    def update_macroform(self, pane, macro):
        i = 0
        for i, variable in enumerate(macro.func_code.co_varnames): 
            self.varlabels[i].SetLabel(variable)
            self.varlabels[i].Show(True)
            self.varentries[i].Show(True)
        for i in xrange(i+1, self.maxvar):
            self.varlabels[i].Hide()
            self.varentries[i].Hide()
        pane.OnToggle(None)
        pane.OnToggle(None)
        
    def make_macroform_content(self, pane):
        self.varlabels = []
        self.varentries = []
        self.macroformsizer = wx.FlexGridSizer(cols=2, hgap=5, vgap=5)
        self.macroformsizer.AddGrowableRow(0)
        #self.macroformsizer.AddGrowableCol(0)
        self.macroformsizer.AddGrowableCol(1)
        for i in xrange(self.maxvar):
            self.varlabels.append(wx.StaticText(pane, -1, "", style=wx.ALIGN_RIGHT))
            self.varentries.append(wx.TextCtrl(pane, -1, ""))
            self.varlabels[-1].Hide()
            self.varentries[-1].Hide()
            self.macroformsizer.Add(self.varlabels[i], 0, wx.EXPAND, 0)
            self.macroformsizer.Add(self.varentries[i], 1, wx.EXPAND, 0)
        self.ok_button = wx.Button(pane, wx.ID_OK)
        self.cancel_button = wx.Button(pane, wx.ID_CANCEL)
        self.macroformsizer.Add(self.cancel_button, 0, wx.EXPAND, 0)
        self.macroformsizer.Add(self.ok_button, 1, wx.EXPAND, 0)
        pane.SetSizer(self.macroformsizer)
        self.GetParent().Layout() 
    
    def make_code_content(self, pane):
        self.CodeTextCtrl = PythonSTC(pane, wx.NewId(), pos=wx.DefaultPosition, size=wx.DefaultSize, 
                                      style=wx.TE_PROCESS_ENTER|wx.TE_PROCESS_TAB|wx.TE_MULTILINE|wx.EXPAND)
        
        self.CodeTextCtrl.SetToolTipString("Enter one python function here.\nThe first line has to begin with def")
        border = wx.FlexGridSizer(cols=0, hgap=5, vgap=5)
        border.AddGrowableCol(0)
        border.AddGrowableRow(0)
        border.Add(self.CodeTextCtrl, 0, wx.EXPAND|wx.FIXED_MINSIZE, 0)
        pane.SetSizer(border)

# end of class MacroEditPanel


class SortedListCtrl(wx.ListCtrl, listmix.ListCtrlAutoWidthMixin):
    """ListCtrl with items sorted in first column"""
    
    def __init__(self, parent, ID, pos=wx.DefaultPosition,
                 size=wx.DefaultSize, style=0):
        wx.ListCtrl.__init__(self, parent, ID, pos, size, style)
        listmix.ListCtrlAutoWidthMixin.__init__(self)

# end of class SortedListCtrl

class PythonSTC(stc.StyledTextCtrl):
    """Editor that highlights Python source code.
    
    Stolen from the wxPython demo.py
    """
    fold_symbols = 2
    
    def __init__(self, parent, ID,
                 pos=wx.DefaultPosition, size=wx.DefaultSize,
                 style=0):
        stc.StyledTextCtrl.__init__(self, parent, ID, pos, size, style)

        self.CmdKeyAssign(ord('B'), stc.STC_SCMOD_CTRL, stc.STC_CMD_ZOOMIN)
        self.CmdKeyAssign(ord('N'), stc.STC_SCMOD_CTRL, stc.STC_CMD_ZOOMOUT)

        self.SetLexer(stc.STC_LEX_PYTHON)
        self.SetKeyWords(0, " ".join(keyword.kwlist))

        self.SetProperty("fold", "1")
        self.SetProperty("tab.timmy.whinge.level", "1")
        self.SetMargins(0,0)

        self.SetViewWhiteSpace(False)
        #self.SetBufferedDraw(False)
        #self.SetViewEOL(True)
        #self.SetEOLMode(stc.STC_EOL_CRLF)
        self.SetUseAntiAliasing(True)
        
        self.SetEdgeMode(stc.STC_EDGE_BACKGROUND)
        self.SetEdgeColumn(78)

        # Setup a margin to hold fold markers
        #self.SetFoldFlags(16)  ###  WHAT IS THIS VALUE?  WHAT ARE THE OTHER FLAGS?  DOES IT MATTER?
        self.SetMarginType(2, stc.STC_MARGIN_SYMBOL)
        self.SetMarginMask(2, stc.STC_MASK_FOLDERS)
        self.SetMarginSensitive(2, True)
        self.SetMarginWidth(2, 12)

        if self.fold_symbols == 0:
            # Arrow pointing right for contracted folders, arrow pointing down for expanded
            self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPEN,    stc.STC_MARK_ARROWDOWN, "black", "black")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDER,        stc.STC_MARK_ARROW, "black", "black")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDERSUB,     stc.STC_MARK_EMPTY, "black", "black")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDERTAIL,    stc.STC_MARK_EMPTY, "black", "black")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDEREND,     stc.STC_MARK_EMPTY,     "white", "black")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPENMID, stc.STC_MARK_EMPTY,     "white", "black")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDERMIDTAIL, stc.STC_MARK_EMPTY,     "white", "black")
            
        elif self.fold_symbols == 1:
            # Plus for contracted folders, minus for expanded
            self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPEN,    stc.STC_MARK_MINUS, "white", "black")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDER,        stc.STC_MARK_PLUS,  "white", "black")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDERSUB,     stc.STC_MARK_EMPTY, "white", "black")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDERTAIL,    stc.STC_MARK_EMPTY, "white", "black")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDEREND,     stc.STC_MARK_EMPTY, "white", "black")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPENMID, stc.STC_MARK_EMPTY, "white", "black")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDERMIDTAIL, stc.STC_MARK_EMPTY, "white", "black")

        elif self.fold_symbols == 2:
            # Like a flattened tree control using circular headers and curved joins
            self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPEN,    stc.STC_MARK_CIRCLEMINUS,          "white", "#404040")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDER,        stc.STC_MARK_CIRCLEPLUS,           "white", "#404040")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDERSUB,     stc.STC_MARK_VLINE,                "white", "#404040")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDERTAIL,    stc.STC_MARK_LCORNERCURVE,         "white", "#404040")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDEREND,     stc.STC_MARK_CIRCLEPLUSCONNECTED,  "white", "#404040")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPENMID, stc.STC_MARK_CIRCLEMINUSCONNECTED, "white", "#404040")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDERMIDTAIL, stc.STC_MARK_TCORNERCURVE,         "white", "#404040")

        elif self.fold_symbols == 3:
            # Like a flattened tree control using square headers
            self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPEN,    stc.STC_MARK_BOXMINUS,          "white", "#808080")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDER,        stc.STC_MARK_BOXPLUS,           "white", "#808080")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDERSUB,     stc.STC_MARK_VLINE,             "white", "#808080")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDERTAIL,    stc.STC_MARK_LCORNER,           "white", "#808080")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDEREND,     stc.STC_MARK_BOXPLUSCONNECTED,  "white", "#808080")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPENMID, stc.STC_MARK_BOXMINUSCONNECTED, "white", "#808080")
            self.MarkerDefine(stc.STC_MARKNUM_FOLDERMIDTAIL, stc.STC_MARK_TCORNER,           "white", "#808080")


        self.Bind(stc.EVT_STC_UPDATEUI, self.OnUpdateUI)
        self.Bind(stc.EVT_STC_MARGINCLICK, self.OnMarginClick)
        self.Bind(wx.EVT_KEY_DOWN, self.OnKeyPressed)

        # Make some styles,  The lexer defines what each style is used for, we
        # just have to define what each style looks like.  This set is adapted from
        # Scintilla sample property files.

        # Global default styles for all languages
        self.StyleSetSpec(stc.STC_STYLE_DEFAULT,     "face:%(helv)s,size:%(size)d" % faces)
        self.StyleClearAll()  # Reset all to be like the default

        # Global default styles for all languages
        self.StyleSetSpec(stc.STC_STYLE_DEFAULT,     "face:%(helv)s,size:%(size)d" % faces)
        self.StyleSetSpec(stc.STC_STYLE_LINENUMBER,  "back:#C0C0C0,face:%(helv)s,size:%(size2)d" % faces)
        self.StyleSetSpec(stc.STC_STYLE_CONTROLCHAR, "face:%(other)s" % faces)
        self.StyleSetSpec(stc.STC_STYLE_BRACELIGHT,  "fore:#FFFFFF,back:#0000FF,bold")
        self.StyleSetSpec(stc.STC_STYLE_BRACEBAD,    "fore:#000000,back:#FF0000,bold")

        # Python styles
        # Default 
        self.StyleSetSpec(stc.STC_P_DEFAULT, "fore:#000000,face:%(helv)s,size:%(size)d" % faces)
        # Comments
        self.StyleSetSpec(stc.STC_P_COMMENTLINE, "fore:#007F00,face:%(other)s,size:%(size)d" % faces)
        # Number
        self.StyleSetSpec(stc.STC_P_NUMBER, "fore:#007F7F,size:%(size)d" % faces)
        # String
        self.StyleSetSpec(stc.STC_P_STRING, "fore:#7F007F,face:%(helv)s,size:%(size)d" % faces)
        # Single quoted string
        self.StyleSetSpec(stc.STC_P_CHARACTER, "fore:#7F007F,face:%(helv)s,size:%(size)d" % faces)
        # Keyword
        self.StyleSetSpec(stc.STC_P_WORD, "fore:#00007F,bold,size:%(size)d" % faces)
        # Triple quotes
        self.StyleSetSpec(stc.STC_P_TRIPLE, "fore:#7F0000,size:%(size)d" % faces)
        # Triple double quotes
        self.StyleSetSpec(stc.STC_P_TRIPLEDOUBLE, "fore:#7F0000,size:%(size)d" % faces)
        # Class name definition
        self.StyleSetSpec(stc.STC_P_CLASSNAME, "fore:#0000FF,bold,underline,size:%(size)d" % faces)
        # Function or method name definition
        self.StyleSetSpec(stc.STC_P_DEFNAME, "fore:#007F7F,bold,size:%(size)d" % faces)
        # Operators
        self.StyleSetSpec(stc.STC_P_OPERATOR, "bold,size:%(size)d" % faces)
        # Identifiers
        self.StyleSetSpec(stc.STC_P_IDENTIFIER, "fore:#000000,face:%(helv)s,size:%(size)d" % faces)
        # Comment-blocks
        self.StyleSetSpec(stc.STC_P_COMMENTBLOCK, "fore:#7F7F7F,size:%(size)d" % faces)
        # End of line where string is not closed
        self.StyleSetSpec(stc.STC_P_STRINGEOL, "fore:#000000,face:%(mono)s,back:#E0C0E0,eol,size:%(size)d" % faces)

        self.SetCaretForeground("BLUE")
        
        self.SetMarginType(1, stc.STC_MARGIN_NUMBER)
        self.SetMarginWidth(1, 30)

        # register some images for use in the AutoComplete box.
        self.RegisterImage(2, 
            wx.ArtProvider.GetBitmap(wx.ART_NEW, size=(16,16)))
        self.RegisterImage(3, 
            wx.ArtProvider.GetBitmap(wx.ART_COPY, size=(16,16)))


    def OnKeyPressed(self, event):
        if self.CallTipActive():
            self.CallTipCancel()
        key = event.GetKeyCode()

        if key == 32 and event.ControlDown():
            pos = self.GetCurrentPos()

            # Tips
            if event.ShiftDown():
                self.CallTipSetBackground("yellow")
                self.CallTipShow(pos, 'lots of of text: blah, blah, blah\n\n'
                                 'show some suff, maybe parameters..\n\n'
                                 'fubar(param1, param2)')
            # Code completion
            else:
                #lst = []
                #for x in range(50000):
                #    lst.append('%05d' % x)
                #st = " ".join(lst)
                #print len(st)
                #self.AutoCompShow(0, st)

                kw = keyword.kwlist[:]
#                kw.append("zzzzzz?2")
#                kw.append("aaaaa?2")
#                kw.append("__init__?3")
#                kw.append("zzaaaaa?2")
#                kw.append("zzbaaaa?2")
#                kw.append("this_is_a_longer_value")
#                #kw.append("this_is_a_much_much_much_much_much_much_much_longer_value")

                kw.sort()  # Python sorts are case sensitive
                self.AutoCompSetIgnoreCase(False)  # so this needs to match

                # Images are specified with a appended "?type"
                for i in range(len(kw)):
                    if kw[i] in keyword.kwlist:
                        kw[i] = kw[i] + "?1"

                self.AutoCompShow(0, " ".join(kw))
        else:
            event.Skip()


    def OnUpdateUI(self, evt):
        # check for matching braces
        braceAtCaret = -1
        braceOpposite = -1
        charBefore = None
        caretPos = self.GetCurrentPos()

        if caretPos > 0:
            charBefore = self.GetCharAt(caretPos - 1)
            styleBefore = self.GetStyleAt(caretPos - 1)

        # check before
        if charBefore and chr(charBefore) in "[]{}()" and styleBefore == stc.STC_P_OPERATOR:
            braceAtCaret = caretPos - 1

        # check after
        if braceAtCaret < 0:
            charAfter = self.GetCharAt(caretPos)
            styleAfter = self.GetStyleAt(caretPos)

            if charAfter and chr(charAfter) in "[]{}()" and styleAfter == stc.STC_P_OPERATOR:
                braceAtCaret = caretPos

        if braceAtCaret >= 0:
            braceOpposite = self.BraceMatch(braceAtCaret)

        if braceAtCaret != -1  and braceOpposite == -1:
            self.BraceBadLight(braceAtCaret)
        else:
            self.BraceHighlight(braceAtCaret, braceOpposite)
            #pt = self.PointFromPosition(braceOpposite)
            #self.Refresh(True, wxRect(pt.x, pt.y, 5,5))
            #print pt
            #self.Refresh(False)


    def OnMarginClick(self, evt):
        # fold and unfold as needed
        if evt.GetMargin() == 2:
            if evt.GetShift() and evt.GetControl():
                self.FoldAll()
            else:
                lineClicked = self.LineFromPosition(evt.GetPosition())

                if self.GetFoldLevel(lineClicked) & stc.STC_FOLDLEVELHEADERFLAG:
                    if evt.GetShift():
                        self.SetFoldExpanded(lineClicked, True)
                        self.Expand(lineClicked, True, True, 1)
                    elif evt.GetControl():
                        if self.GetFoldExpanded(lineClicked):
                            self.SetFoldExpanded(lineClicked, False)
                            self.Expand(lineClicked, False, True, 0)
                        else:
                            self.SetFoldExpanded(lineClicked, True)
                            self.Expand(lineClicked, True, True, 100)
                    else:
                        self.ToggleFold(lineClicked)


    def FoldAll(self):
        lineCount = self.GetLineCount()
        expanding = True

        # find out if we are folding or unfolding
        for lineNum in range(lineCount):
            if self.GetFoldLevel(lineNum) & stc.STC_FOLDLEVELHEADERFLAG:
                expanding = not self.GetFoldExpanded(lineNum)
                break

        lineNum = 0

        while lineNum < lineCount:
            level = self.GetFoldLevel(lineNum)
            if level & stc.STC_FOLDLEVELHEADERFLAG and \
               (level & stc.STC_FOLDLEVELNUMBERMASK) == stc.STC_FOLDLEVELBASE:

                if expanding:
                    self.SetFoldExpanded(lineNum, True)
                    lineNum = self.Expand(lineNum, True)
                    lineNum = lineNum - 1
                else:
                    lastChild = self.GetLastChild(lineNum, -1)
                    self.SetFoldExpanded(lineNum, False)

                    if lastChild > lineNum:
                        self.HideLines(lineNum+1, lastChild)

            lineNum = lineNum + 1



    def Expand(self, line, doExpand, force=False, visLevels=0, level=-1):
        lastChild = self.GetLastChild(line, level)
        line = line + 1

        while line <= lastChild:
            if force:
                if visLevels > 0:
                    self.ShowLines(line, line)
                else:
                    self.HideLines(line, line)
            else:
                if doExpand:
                    self.ShowLines(line, line)

            if level == -1:
                level = self.GetFoldLevel(line)

            if level & stc.STC_FOLDLEVELHEADERFLAG:
                if force:
                    if visLevels > 1:
                        self.SetFoldExpanded(line, True)
                    else:
                        self.SetFoldExpanded(line, False)

                    line = self.Expand(line, doExpand, force, visLevels-1)

                else:
                    if doExpand and self.GetFoldExpanded(line):
                        line = self.Expand(line, True, force, visLevels-1)
                    else:
                        line = self.Expand(line, False, force, visLevels-1)
            else:
                line = line + 1

        return line
        
# end of class PythonSTC



#!/usr/bin/env python

from distutils.core import setup

setup(name='pyspread',
      version='0.0.10',
      description='A spreadsheet that accepts a pure python expression in each cell.',
      license='GPL v3 :: GNU General Public License',
      classifiers=[ 'Development Status :: 3 - Alpha',
                    'Intended Audience :: End Users/Desktop',
      ],
      author='Martin Manns',
      author_email='mmanns@gmx.net',
      url='http://sourceforge.net/projects/pyspread/',
      requires=['numpy (>=1.1)', 'wx (>=2.7)'],
      scripts=['pyspread/pyspread.py'],
      packages=['pyspread'],
      package_dir={'pyspread': 'pyspread'},
#      py_modules=['pyspread', 'config', '_choicebars', '_datastructures',\
#                  '_interfaces', '_widgets', '_dialogs'],
      package_data={'pyspread': ['icons/*.png', 'icons/actions/*.png', \
                         'examples/*', \
                         'doc/manual.html', 'README', 'COPYING']},
)

import distutils.sysconfig
try:
    pthfile = open(distutils.sysconfig.get_python_lib()+"/pyspread.pth",'w')
    pthfile.write("pyspread")
    pthfile.close()
except: print 'Creation of ' + distutils.sysconfig.get_python_lib() + ' pyspread.pth failed.'
